"""Base class for the code execution backend."""

import asyncio
import glob
import io
import json
import os
import resource
import tarfile
import time
from abc import abstractmethod
from typing import TYPE_CHECKING, List, Optional

import aiofiles
import pyarrow
from pydantic import BaseModel, ValidationError

from bespokelabs.curator.code_executor.code_formatter import CodeFormatter
from bespokelabs.curator.code_executor.tracker import CodeExecutionStatusTracker
from bespokelabs.curator.code_executor.types import CodeAPIRequest, CodeExecutionOutput, CodeExecutionRequest, CodeExecutionResponse
from bespokelabs.curator.constants import _CACHE_MSG
from bespokelabs.curator.file_utilities import count_lines
from bespokelabs.curator.log import logger
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop

if TYPE_CHECKING:
    from datasets import Dataset


class BaseCodeExecutionBackend:
    """Base class for all request processors.

    Provides core functionality for processing requests through LLM APIs, including:
    - File descriptor limit management
    - Request file creation and caching
    - Response processing and dataset generation
    - Error handling and validation
    """

    def __init__(self, config: dict):
        """Initialize the request processor.

        Args:
            config: Configuration object containing processing parameters
        """
        # Increase the number of open file descriptors to avoid "Too many open files" errors
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        desired_limit = min(10_000_000, hard)
        logger.debug(f"Adjusting file descriptor limit from {soft} to {desired_limit} (hard limit: {hard})")
        resource.setrlimit(resource.RLIMIT_NOFILE, (desired_limit, hard))
        self.config = config

        self.manual_max_requests_per_minute = config.max_requests_per_minute
        self.default_max_requests_per_minute = 10
        # The rich.Console used for the status tracker, only set for testing
        self._tracker_console = None

    @property
    @abstractmethod
    def backend(self) -> str:
        """Backend property."""
        return "base"

    @abstractmethod
    async def execute_request(self, request: CodeAPIRequest) -> CodeExecutionResponse:
        """Execute a single request."""
        pass

    @abstractmethod
    def requests_to_responses(self, execution_request_files: list[str]) -> None:
        """Process request files and generate responses.

        Args:
            execution_request_files: List of paths to request files to process
        """
        for request_file in execution_request_files:
            response_file = request_file.replace("requests_", "responses_")
            run_in_event_loop(
                self.process_requests_from_file(
                    execution_request_filepath=request_file,
                    response_file=response_file,
                )
            )

    async def process_requests_from_file(
        self,
        execution_request_filepath: str,
        response_file: str,
    ) -> None:
        """Processes API requests in parallel, throttling to stay under rate limits.

        Args:
            execution_request_filepath: Path to file containing requests
            response_file: Path where the response data will be saved
        """
        # Initialize trackers
        queue_of_requests_to_retry: asyncio.Queue[CodeExecutionRequest] = asyncio.Queue()
        status_tracker = CodeExecutionStatusTracker()

        # Get rate limits
        status_tracker.max_requests_per_minute = self.manual_max_requests_per_minute

        # Resume if a response file exists
        completed_request_ids = self.validate_existing_response_file(response_file)

        # Count total requests
        status_tracker.num_tasks_already_completed = len(completed_request_ids)
        status_tracker.total_requests = self.total_requests
        status_tracker.start_tracker(self._tracker_console)

        # Use higher connector limit for better throughput
        async with aiofiles.open(execution_request_filepath) as file:
            pending_requests = []

            async for line in file:
                execution_request = CodeExecutionRequest.model_validate_json(line)

                if execution_request.original_row_idx in completed_request_ids:
                    continue

                request = CodeAPIRequest(
                    task_id=execution_request.original_row_idx,
                    execution_request=execution_request,
                    attempts_left=self.config.max_retries,
                    code_formatter=self.code_formatter,
                )

                # print(f"Processing request {request.task_id}")

                # while not status_tracker.has_capacity():
                # await asyncio.sleep(0.1)

                status_tracker.consume_capacity()

                # Wait for rate limits cool down if needed
                # todo: implement this
                # await self.cool_down_if_rate_limit_error(status_tracker)

                task = asyncio.create_task(
                    self.handle_single_request_with_retries(
                        request=request,
                        retry_queue=queue_of_requests_to_retry,
                        response_file=response_file,
                        status_tracker=status_tracker,
                    )
                )

                pending_requests.append(task)

                # status_tracker.num_tasks_started += 1
                # status_tracker.num_tasks_in_progress += 1

            # Wait for all tasks to complete
            if pending_requests:
                await asyncio.gather(*pending_requests)

            # Process any remaining retries in the queue
            pending_retries = set()
            while not queue_of_requests_to_retry.empty() or pending_retries:
                # Process new items from the queue if we have capacity
                if not queue_of_requests_to_retry.empty():
                    retry_request = await queue_of_requests_to_retry.get()
                    attempt_number = self.config.max_retries - retry_request.attempts_left
                    logger.debug(
                        f"Retrying request {retry_request.task_id} "
                        f"(attempt #{attempt_number} of {self.config.max_retries})"
                        f"Previous errors: {retry_request.result}"
                    )

                    # Wait for capacity if needed
                    # while not status_tracker.has_capacity():
                    # await asyncio.sleep(0.1)

                    # Consume capacity before making request
                    # status_tracker.consume_capacity()

                    task = asyncio.create_task(
                        self.handle_single_request_with_retries(
                            request=retry_request,
                            retry_queue=queue_of_requests_to_retry,
                            response_file=response_file,
                            status_tracker=status_tracker,
                        )
                    )
                    pending_retries.add(task)

                # Wait for some tasks to complete
                if pending_retries:
                    done, pending_retries = await asyncio.wait(pending_retries, timeout=0.1)

        status_tracker.stop_tracker()

        # Log final status
        logger.info(f"Processing complete. Results saved to {response_file}")
        logger.info(f"Status tracker: {status_tracker}")

        if status_tracker.num_tasks_failed > 0:
            logger.warning(f"{status_tracker.num_tasks_failed} / {status_tracker.num_tasks_started} requests failed. Errors logged to {response_file}.")

    async def handle_single_request_with_retries(
        self,
        request: CodeAPIRequest,
        retry_queue: asyncio.Queue,
        response_file: str,
        status_tracker: CodeExecutionStatusTracker,
    ) -> None:
        """Common wrapper for handling a single request with error handling and retries.

        This method implements the common try/except logic and retry mechanism,
        while delegating the actual API call to call_single_request.

        Args:
            request: The request to process
            session: Async HTTP session
            retry_queue: Queue for failed requests
            response_file: Path where the response data will be saved
            status_tracker: Tracks request status
        """
        try:
            execution_output = await self.execute_request(request)

            execution_response = CodeExecutionResponse(
                code_api_request=request,
                exec_output=execution_output,
            ).model_dump()

            await self.append_execution_response(execution_response, response_file)

            status_tracker.num_tasks_in_progress -= 1
            status_tracker.num_tasks_succeeded += 1
            status_tracker.update_stats()

        except Exception as e:
            status_tracker.num_other_errors += 1

            request.result.append(e)

            if request.attempts_left > 0:
                request.attempts_left -= 1
                logger.warning(
                    f"Encountered '{e.__class__.__name__}: {e}' during attempt "
                    f"{self.config.max_retries - request.attempts_left} of {self.config.max_retries} "
                    f"while processing request {request.task_id}"
                )
                retry_queue.put_nowait(request)
            else:
                logger.error(
                    f"Request {request.task_id} failed permanently after exhausting all {self.config.max_retries} retry attempts. "
                    f"Errors: {[str(e) for e in request.result]}"
                )

                execution_response = CodeExecutionResponse(
                    code_api_request=request,
                    exec_output=CodeExecutionOutput(
                        message="Error",
                        error="\n".join([str(e) for e in request.result]),
                    ),
                ).model_dump()

                await self.append_execution_response(execution_response, response_file)

    def run(
        self,
        dataset: "Dataset",
        working_dir: str,
        code_formatter: CodeFormatter,
        all_func_hash_hash: str,
    ) -> "Dataset":
        """Uses the API to completing the specific map by calling the LLM.

        Args:
            dataset: Dataset that is being mapped over
            working_dir: Working directory to save files (requests.jsonl, responses.jsonl, dataset.arrow)
            code_formatter: Code formatter to use
            all_func_hash_hash: Hash identifying the dataset version

        Returns:
            Dataset: Completed dataset with responses

        Raises:
            ValueError: If model doesn't support structured output but it's requested
        """
        self.code = code_formatter.code
        self.code_input = code_formatter.code_input
        self.code_output = code_formatter.code_output
        self.working_dir = working_dir
        self.total_requests = len(dataset) if dataset is not None else 1
        # load from already completed dataset

        output_dataset = self.attempt_loading_cached_dataset(all_func_hash_hash)
        if output_dataset is not None:
            return output_dataset

        self.code_formatter = code_formatter

        execution_request_files = self.create_request_files(dataset)

        self.requests_to_responses(execution_request_files)

        return self.create_dataset_files(all_func_hash_hash)

    def _verify_existing_request_files(self, dataset: Optional["Dataset"]) -> List[int]:
        """Verify integrity of the cache and identify files needing regeneration.

        Checks that each request file has associated metadata and correct number of rows.

        Args:
            dataset: The dataset to create requests from

        Returns:
            List of indices for request files that need to be regenerated
        """
        expected_num_files = 1

        try:
            incomplete_files = []
            for i in range(expected_num_files):
                req_f = os.path.join(self.working_dir, f"requests_{i}.jsonl")
                meta_f = os.path.join(self.working_dir, f"metadata_{i}.json")

                if not os.path.exists(req_f):
                    incomplete_files.append(i)
                    continue

                if not os.path.exists(meta_f):
                    logger.warning(f"Cache missing metadata file {meta_f} for request file {req_f}")
                    incomplete_files.append(i)
                    continue

                with open(req_f, "r") as f:
                    data = f.read()
                num_jobs = len(data.splitlines())

                with open(meta_f, "r") as f:
                    metadata = json.load(f)

                expected_num_jobs = metadata["num_jobs"]
                if num_jobs != expected_num_jobs:
                    logger.warning(f"Request file {req_f} has {num_jobs} jobs, but metadata file {meta_f} has {expected_num_jobs} jobs")
                    incomplete_files.append(i)
            return incomplete_files

        except Exception as e:
            logger.warning(f"Cache verification failed due to {e} - regenerating all request files.")
            incomplete_files = list(range(expected_num_files))
            return incomplete_files

    def create_request_files(self, dataset: Optional["Dataset"]) -> list[str]:
        """Creates request files if they don't exist or uses existing ones.

        Args:
            dataset: The dataset to be processed

        Returns:
            List of paths to the request files
        """
        os.makedirs(self.working_dir, exist_ok=True)
        request_files = glob.glob(os.path.join(self.working_dir, "requests_*.jsonl"))

        # By default use existing requests in working_dir
        incomplete_files = self._verify_existing_request_files(dataset)

        if len(incomplete_files) == 0:
            logger.info(f"Using cached requests. {_CACHE_MSG}")
            # count existing jobs in file and print first job
            with open(request_files[0], "r") as f:
                # Count lines and store first job
                first_job = None
                num_jobs = 0
                for i, line in enumerate(f):
                    if i == 0:
                        first_job = json.loads(line)
                    num_jobs = i + 1

                if num_jobs > 0:
                    logger.debug(
                        f"There are {num_jobs} existing requests in {request_files[0]}\n"
                        f"Example request in {request_files[0]}:\n{json.dumps(first_job, default=str, indent=2)}"
                    )
                    return request_files

        # Create new requests file
        logger.info(f"Preparing request file(s) in {self.working_dir}")
        request_file = os.path.join(self.working_dir, "requests_0.jsonl")
        request_files = [request_file]

        metadata_file = os.path.join(self.working_dir, "metadata_0.json")

        if dataset is None:
            raise ValueError("Dataset is empty")

        run_in_event_loop(self.acreate_request_file(dataset, request_file, metadata_file))

        return request_files

    async def acreate_request_file(
        self,
        dataset: "Dataset",
        request_file: str,
        metadata_file: str,
        start_idx: int = 0,
    ) -> None:
        """Asynchronously create a request file and its metadata.

        Args:
            dataset: Dataset to create requests from
            request_file: Path to save request file
            metadata_file: Path to save metadata file
            start_idx: Starting index in dataset for this batch
        """
        end_idx = len(dataset)

        async with aiofiles.open(request_file, "w") as f:
            for idx, dataset_row in enumerate(dataset):
                dataset_row_idx = idx + start_idx
                # Get the generic request from the map function
                execution_directory = os.path.join(self.working_dir, "code_execution", str(dataset_row_idx))
                request = self.code_formatter.create_code_execution_request(dataset_row, dataset_row_idx, execution_directory)
                await f.write(json.dumps(request.model_dump(), default=str) + "\n")

        num_requests = end_idx - start_idx
        metadata_dict = {"num_jobs": num_requests}
        async with aiofiles.open(metadata_file, "w") as f:
            await f.write(json.dumps(metadata_dict, indent=4) + "\n")

        logger.info(f"Wrote {num_requests} requests to {request_file}.")

    def attempt_loading_cached_dataset(self, all_func_hash_hash: str) -> Optional["Dataset"]:
        """Attempt to load a cached dataset file.

        Args:
            all_func_hash_hash: Hash identifying the specific dataset

        Returns:
            Cached dataset if available and valid, None otherwise
        """
        dataset_file = os.path.join(self.working_dir, f"{all_func_hash_hash}.arrow")
        if os.path.exists(dataset_file):
            logger.debug(f"Loading dataset from {dataset_file}")
            try:
                logger.info(f"Using cached output dataset. {_CACHE_MSG}")
                return self._load_from_dataset_file(dataset_file)
            except pyarrow.lib.ArrowInvalid:
                os.remove(dataset_file)
                logger.warning(
                    f"Failed to load dataset from {dataset_file}, "
                    "which was likely corrupted by a failed previous run. "
                    "Deleted file and attempting to regenerate dataset from cached LLM responses."
                )

    def create_dataset_files(
        self,
        all_func_hash_hash: str,
    ) -> "Dataset":
        """Creates dataset from response files.

        Args:
            all_func_hash_hash: Hash identifying the dataset version

        Returns:
            Dataset containing processed responses

        Raises:
            ValueError: If no responses found or processing fails
        """
        responses_files = glob.glob(os.path.join(self.working_dir, "responses_*.jsonl"))
        if len(responses_files) == 0:
            raise ValueError(f"No responses files found in {self.working_dir}")

        error_help = (
            "Please check your `parse_func` is returning a valid row (dict) "
            "or list of rows (list of dicts) and re-run. "
            "Dataset will be regenerated from cached LLM responses."
        )

        # Process all response files
        total_responses_count = 0
        failed_responses_count = 0
        error_sample = []
        dataset_file = os.path.join(self.working_dir, f"{all_func_hash_hash}.arrow")
        from datasets.arrow_writer import ArrowWriter

        with ArrowWriter(path=dataset_file) as writer:
            for responses_file in responses_files:
                with open(responses_file, "r") as f_in:
                    for execution_response_string in f_in:
                        total_responses_count += 1

                        response = CodeExecutionResponse.model_validate_json(execution_response_string)

                        if self.code_formatter.code_output:
                            try:
                                dataset_rows = self.code_formatter.code_output(
                                    response.code_api_request.execution_request.original_row,
                                    response.exec_output,
                                )
                            except Exception as e:
                                logger.error(f"Exception raised in your `parse_func`. {error_help}")
                                os.remove(dataset_file)
                                raise e
                            if not isinstance(dataset_rows, list):
                                dataset_rows = [dataset_rows]
                        else:
                            raise ValueError("code_output is not implemented")

                        for row in dataset_rows:
                            if isinstance(row, BaseModel):
                                row = row.model_dump()

                            if not isinstance(row, dict):
                                os.remove(dataset_file)
                                raise ValueError(
                                    f"Got invalid row {row} of type {type(row)} from `parse_func`. This should be type <class 'dict'>. {error_help}"
                                )
                            if not row:
                                os.remove(dataset_file)
                                raise ValueError(f"Got empty row {row} from `parse_func`. {error_help}")
                            # Add the original row index to the row so that we can sort by it later.
                            row["__original_row_idx"] = response.code_api_request.execution_request.original_row_idx
                            writer.write(row)

            logger.info(f"Read {total_responses_count} responses.")
            error_sample_str = "\n".join(error_sample)
            error_sample_msg = f"Sample of the first {len(error_sample)} errors encountered: \n {error_sample_str}"
            if failed_responses_count == total_responses_count:
                writer.write({"error": "All requests failed"})
                writer.finalize()
                os.remove(dataset_file)
                raise ValueError(f"All requests failed. {error_sample_msg}")
            else:
                logger.info("Finalizing writer")
                writer.finalize()

            if failed_responses_count > 0:
                logger.warning(f"{failed_responses_count} requests failed.")
                if self.config.require_all_responses:
                    os.remove(dataset_file)
                    raise ValueError(f"Some requests failed and require_all_responses is True. {error_sample_msg}")

            # number of responses matches number of requests
            request_files = glob.glob(os.path.join(self.working_dir, "requests_*.jsonl"))
            n_requests = 0
            for request_file in request_files:
                n_requests += count_lines(request_file)

            if n_requests != total_responses_count:
                logger.warning(
                    f"{n_requests - total_responses_count} requests do not have responses. "
                    f"n_requests is {n_requests} and n_responses is {total_responses_count}"
                )
                if self.config.require_all_responses:
                    os.remove(dataset_file)
                    raise ValueError("Some requests do not have responses and require_all_responses is True.")

        return self._load_from_dataset_file(dataset_file)

    def _load_from_dataset_file(self, dataset_file: str) -> "Dataset":
        from datasets import Dataset

        print(f"Loading dataset from {dataset_file}")
        d = Dataset.from_file(dataset_file)
        d = d.sort("__original_row_idx")
        d = d.remove_columns("__original_row_idx")
        return d

    def validate_existing_response_file(self, response_file: str) -> set[int]:
        """Parse an existing response file to identify completed requests and removes failed requests.

        Args:
            response_file: Path to the response file to parse

        Returns:
            set[int]: Set of completed request IDs that were already successfully processed
        """
        completed_request_ids = set()
        failed_request_ids = set()

        if os.path.exists(response_file):
            logger.info(f"Resuming progress by reading existing file: {response_file}")
            logger.debug(f"Removing all failed requests from {response_file} so they can be retried")
            temp_filepath = response_file + ".temp"

            parsing_error_responses = 0
            with open(response_file, "r") as input_file, open(temp_filepath, "w") as output_file:
                for line in input_file:
                    try:
                        response = CodeExecutionResponse.model_validate_json(line)
                    except (json.JSONDecodeError, ValidationError):
                        logger.warning("Skipping response due to error parsing line")
                        parsing_error_responses += 1
                        continue
                    row_id = response.code_api_request.execution_request.original_row_idx
                    if response.exec_output.error:
                        logger.debug(f"Request {row_id} previously failed due to errors: {response.exec_output.error}, removing from output and will retry")
                        failed_request_ids.add(row_id)
                    elif response.exec_output.message is None:
                        logger.debug(f"Request {row_id} previously failed due to no response. Removing from output and will retry.")
                        failed_request_ids.add(row_id)
                    else:
                        completed_request_ids.add(row_id)
                        output_file.write(line)

            logger.info(
                f"Found {len(completed_request_ids)} successful requests and {len(failed_request_ids)} "
                f"previously failed requests and {parsing_error_responses} parsing errors in {response_file}"
            )
            os.replace(temp_filepath, response_file)

        return completed_request_ids

    def read_metadata_file(self, request_file: str) -> int:
        """Read the number of jobs from the metadata file.

        Args:
            request_file: Path to the request file to get metadata for

        Returns:
            int: Number of total batch requests

        Raises:
            ValueError: If metadata file is missing or invalid
        """
        metadata_file = request_file.replace("requests_", "metadata_").replace(".jsonl", ".json")

        if not os.path.exists(metadata_file):
            raise ValueError(f"Metadata file not found: {metadata_file}")

        try:
            with open(metadata_file, "r") as f:
                metadata = json.load(f)
                return metadata
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in metadata file: {metadata_file}. Delete cache directory 'rm -rf {self.working_dir}' and try again.") from e

    async def cool_down_if_rate_limit_error(self, status_tracker: CodeExecutionStatusTracker) -> None:
        """Pause processing if a rate limit error is detected.

        Args:
            status_tracker: Tracker containing rate limit status
        """
        seconds_to_pause_on_rate_limit = self.config.seconds_to_pause_on_rate_limit
        seconds_since_rate_limit_error = time.time() - status_tracker.time_of_last_rate_limit_error
        remaining_seconds_to_pause = seconds_to_pause_on_rate_limit - seconds_since_rate_limit_error
        if remaining_seconds_to_pause > 0:
            logger.warn(f"Pausing for {int(remaining_seconds_to_pause)} seconds")
            await asyncio.sleep(remaining_seconds_to_pause)

    async def append_execution_response(self, data: dict, filename: str) -> None:
        """Append a response to a jsonl file with async file operations.

        Args:
            data: Response data to append
            filename: File to append to
        """
        json_string = json.dumps(data, default=str)
        async with aiofiles.open(filename, "a") as f:
            await f.write(json_string + "\n")
        logger.debug(f"Successfully appended response to {filename}")

    @classmethod
    def _create_temp_file(cls, content: str, execution_directory: str) -> str:
        """Create a temporary file with the given content.

        Args:
            content: Content to write to temp file
            execution_directory: Directory to create the temp file in

        Returns:
            Path to the created temp file
        """
        # create execution directory if it doesn't exist
        os.makedirs(execution_directory, exist_ok=True)

        temp_file_path = os.path.join(execution_directory, "program.py")
        with open(temp_file_path, "w", encoding="utf-8") as temp_file:
            temp_file.write(content)
        return temp_file_path

    @classmethod
    def _get_created_files(cls, program_dir: str) -> bytes:
        """Get any files created during code execution.

        Args:
            program_dir: Directory containing the executed program and created files

        Returns:
            Bytes containing a zip archive of any created files
        """
        # Create a BytesIO object to store the zip data
        tar_buffer = io.BytesIO()

        # Create a zip archive containing all files in program_dir
        with tarfile.open(fileobj=tar_buffer, mode="w") as tar:
            # Add all files in program_dir to the archive
            for filename in os.listdir(program_dir):
                file_path = os.path.join(program_dir, filename)
                if os.path.isfile(file_path):
                    tar.add(file_path, arcname=filename)

        # Get the bytes from the buffer
        tar_buffer.seek(0)
        return str(tar_buffer.getvalue())



"""Defaults used across curator."""

BATCH_REQUEST_ID_TAG = "custom_id"
_CURATOR_DEFAULT_CACHE_DIR = "~/.cache/curator"
_DEFAULT_CACHE_DIR = "~/.cache"
BASE_CLIENT_URL = "https://api.bespokelabs.ai/v0/viewer"
PUBLIC_CURATOR_VIEWER_HOME_URL = "https://curator.bespokelabs.ai"
PUBLIC_CURATOR_VIEWER_DATASET_URL = PUBLIC_CURATOR_VIEWER_HOME_URL + "/datasets"
_INTERNAL_PROMPT_KEY = "__internal_prompt"
_CACHE_MSG = (
    "If you want to regenerate the dataset, disable or delete the cache.\n See "
    "https://docs.bespokelabs.ai/bespoke-curator/tutorials/automatic-recovery-and-caching#disable-caching "
    "for more information."
)



"""Docker Code Execution Backend."""

import asyncio
import io
import multiprocessing
import os
import tarfile
import tempfile
from concurrent.futures import ProcessPoolExecutor
from functools import partial

import docker
from docker.errors import DockerException

from bespokelabs.curator.code_executor.code_execution_backend.base_backend import BaseCodeExecutionBackend
from bespokelabs.curator.code_executor.types import CodeAPIRequest, CodeExecutionOutput, CodeExecutionRequestParams, CodeExecutionResponse
from bespokelabs.curator.log import logger


class DockerCodeExecutionBackend(BaseCodeExecutionBackend):
    """Docker-based code execution backend."""

    PYTHON_IMAGE = "python:3.11-slim"
    WORKSPACE_DIR = "/workspace"

    def __init__(self, config):
        """Initialize the backend."""
        super().__init__(config)
        self.config = config
        self.PYTHON_IMAGE = config.docker_image if config.docker_image else self.PYTHON_IMAGE
        self.client = docker.from_env()

        # Initialize process pool for parallel execution
        self.max_workers = getattr(config, "max_workers", multiprocessing.cpu_count())
        self.process_pool = ProcessPoolExecutor(max_workers=self.max_workers)

        logger.debug(f"Initialized Docker backend with {self.max_workers} workers")

    def _ensure_image(self):
        """Ensure required Docker image is pulled."""
        try:
            self.client.images.get(self.PYTHON_IMAGE)
        except docker.errors.ImageNotFound:
            logger.info(f"Pulling {self.PYTHON_IMAGE} image...")
            self.client.images.pull(self.PYTHON_IMAGE)
        except DockerException as e:
            raise RuntimeError(f"Error ensuring Docker image: {str(e)}") from e

    async def execute_request(self, request: CodeAPIRequest) -> CodeExecutionResponse:
        """Execute a single request in a Docker container using a process pool."""
        return await self.execute_standard_input_request(
            request.execution_request.code, request.execution_request.code_input, request.execution_request.execution_params
        )

    @classmethod
    def _create_temp_file(cls, content: str) -> str:
        """Create a temporary file with the given content."""
        with tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as temp_file:
            temp_file.write(content)
            return temp_file.name

    def _create_tar_archive(self, data: bytes, filename: str) -> io.BytesIO:
        """Create a tar archive containing a single file."""
        tarstream = io.BytesIO()
        with tarfile.TarFile(fileobj=tarstream, mode="w") as tar:
            tarinfo = tarfile.TarInfo(name=filename)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        tarstream.seek(0)
        return tarstream

    def _get_container_file(self, container, file_path: str) -> str:
        """Get content of a file from a Docker container."""
        try:
            bits, _ = container.get_archive(file_path)
            bio = io.BytesIO()
            for chunk in bits:
                bio.write(chunk)
            bio.seek(0)

            with tarfile.open(fileobj=bio) as tar:
                member = tar.next()
                f = tar.extractfile(member)
                if f:
                    return f.read().decode("utf-8")
                return ""

        except docker.errors.NotFound:
            return ""

    def _setup_volume_with_files(self, volume, code_file: str, input_file: str):
        """Set up a volume with code and input files."""
        setup_container = self.client.containers.create(
            self.PYTHON_IMAGE, command=["sleep", "5"], volumes={volume.name: {"bind": self.WORKSPACE_DIR, "mode": "rw"}}
        )

        try:
            setup_container.start()
            for src, dest in [(code_file, "program.py"), (input_file, "input.txt")]:
                with open(src, "rb") as f:
                    data = f.read()
                    tarstream = self._create_tar_archive(data, dest)
                    setup_container.put_archive(self.WORKSPACE_DIR, tarstream)
        finally:
            setup_container.stop()
            setup_container.remove()

    def _get_created_files(self, container: docker.models.containers.Container, directory: str) -> str:
        """Get any files created during code execution."""
        # Get all files in the directory as a tar archive
        bits, _ = container.get_archive(directory, encode_stream=True)
        bio = io.BytesIO()
        for chunk in bits:
            bio.write(chunk)
        bio.seek(0)
        return str(bio.getvalue())

    def _create_execution_container(self, volume_name: str) -> docker.models.containers.Container:
        """Create the main execution container."""
        # TODO: Add a name to the container and volume so that they are unique for a each row
        return self.client.containers.create(
            self.PYTHON_IMAGE,
            command=[  # noqa: E501
                f"bash -c python {self.WORKSPACE_DIR}/program.py < {self.WORKSPACE_DIR}/input.txt > {self.WORKSPACE_DIR}/output.txt 2> {self.WORKSPACE_DIR}/error.txt; echo $? > {self.WORKSPACE_DIR}/exit_code.txt"  # noqa: E501
            ],
            volumes={volume_name: {"bind": self.WORKSPACE_DIR, "mode": "rw"}},
            working_dir=self.WORKSPACE_DIR,
            entrypoint=["/bin/sh", "-c"],
            user="root",
        )

    def _handle_execution_result(self, container, status_code: int) -> CodeExecutionOutput:
        """Handle the execution result and create appropriate output."""
        stdout = self._get_container_file(container, f"{self.WORKSPACE_DIR}/output.txt")
        stderr = self._get_container_file(container, f"{self.WORKSPACE_DIR}/error.txt")
        files = self._get_created_files(container, self.WORKSPACE_DIR)

        if status_code == 0:
            return CodeExecutionOutput(
                message="success",
                stdout=stdout,
                stderr=stderr,
                files=files,
            )

        # Create a more descriptive error message
        error_message = f"Program exited with status code {status_code}"
        if stderr:
            error_message = f"{error_message}\n\nError details:\n{stderr}"

        return CodeExecutionOutput(
            message="error",
            error=error_message,
            stdout=stdout,
            stderr=stderr,
            files=files,
        )

    async def execute_standard_input_request(self, code: str, code_input: str, execution_params: CodeExecutionRequestParams) -> CodeExecutionOutput:
        """Execute code in Docker container using a worker from the process pool."""
        # Create a function that can be executed in a separate process
        execution_func = partial(
            self._execute_in_process, code=code, code_input=code_input, timeout=120, python_image=self.PYTHON_IMAGE, workspace_dir=self.WORKSPACE_DIR
        )

        # Submit the task to the process pool
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(self.process_pool, execution_func)

        return result

    @staticmethod
    def _execute_in_process(code: str, code_input: str, timeout: int, python_image: str, workspace_dir: str) -> CodeExecutionOutput:
        """Static method to execute code in a separate process."""
        try:
            # Create a new Docker client in this process
            client = docker.from_env()

            # Create temporary files
            with (
                tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as code_file,
                tempfile.NamedTemporaryFile(delete=False, mode="w", encoding="utf-8") as input_file,
            ):
                code_file.write(code)
                code_file_path = code_file.name

                input_file.write(code_input)
                input_file_path = input_file.name

            # Ensure image exists
            try:
                client.images.get(python_image)
            except docker.errors.ImageNotFound:
                client.images.pull(python_image)
            except DockerException as e:
                raise RuntimeError(f"Error ensuring Docker image: {str(e)}") from e

            container = None
            volume = None

            try:
                volume = client.volumes.create()

                # Set up volume with files
                setup_container = client.containers.create(python_image, command=["sleep", "5"], volumes={volume.name: {"bind": workspace_dir, "mode": "rw"}})

                try:
                    setup_container.start()
                    for src, dest in [(code_file_path, "program.py"), (input_file_path, "input.txt")]:
                        with open(src, "rb") as f:
                            data = f.read()
                            tarstream = DockerCodeExecutionBackend._create_tar_archive_static(data, dest)
                            setup_container.put_archive(workspace_dir, tarstream)
                finally:
                    setup_container.stop()
                    setup_container.remove()

                # Create and run execution container
                cmd = (
                    f"echo 'DEBUG: Starting container execution' && "
                    f"echo 'DEBUG: Workspace contents:' && ls -la {workspace_dir} && "
                    f"python {workspace_dir}/program.py < {workspace_dir}/input.txt > {workspace_dir}/output.txt 2> {workspace_dir}/error.txt; "
                    f"exit_code=$?; "
                    f"echo $exit_code > {workspace_dir}/exit_code.txt; "
                    f"echo 'DEBUG: Execution finished with code:' $exit_code; "
                    f"exit $exit_code"
                )

                container = client.containers.create(
                    python_image,
                    command=[cmd],
                    volumes={volume.name: {"bind": workspace_dir, "mode": "rw"}},
                    working_dir=workspace_dir,
                    entrypoint=["/bin/bash", "-c"],
                    user="root",
                )

                container.start()
                result = container.wait(timeout=timeout)

                # Handle results
                return DockerCodeExecutionBackend._handle_execution_result_static(container, result["StatusCode"], workspace_dir)

            except docker.errors.ContainerError as e:
                return CodeExecutionOutput(message="error", error=f"Docker container error: {str(e)}\n\nExit code: {e.exit_status}\n\nStderr: {e.stderr}")
            except docker.errors.ImageNotFound as e:
                return CodeExecutionOutput(message="error", error=f"Docker image not found: {str(e)}")
            except docker.errors.APIError as e:
                return CodeExecutionOutput(message="error", error=f"Docker API error: {str(e)}")
            except Exception as e:
                return CodeExecutionOutput(message="error", error=f"Error during code execution: {str(e)}\n\nType: {type(e).__name__}")
            finally:
                # Clean up
                try:
                    if container:
                        logs = container.logs(stdout=True, stderr=True).decode("utf-8", errors="replace")
                        print(logs)
                        # container.remove(force=True)
                    # if volume:
                    # volume.remove(force=True)
                    for f in [code_file_path, input_file_path]:
                        if os.path.exists(f):
                            os.unlink(f)
                    client.close()
                except Exception:
                    pass

        except Exception as e:
            return CodeExecutionOutput(message="error", error=f"Process execution error: {str(e)}")

    @staticmethod
    def _create_tar_archive_static(data: bytes, filename: str) -> io.BytesIO:
        """Static version of _create_tar_archive for use in processes."""
        tarstream = io.BytesIO()
        with tarfile.TarFile(fileobj=tarstream, mode="w") as tar:
            tarinfo = tarfile.TarInfo(name=filename)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        tarstream.seek(0)
        return tarstream

    @staticmethod
    def _handle_execution_result_static(container, status_code: int, workspace_dir: str) -> CodeExecutionOutput:
        """Static version of _handle_execution_result for use in processes."""
        stdout = DockerCodeExecutionBackend._get_container_file_static(container, f"{workspace_dir}/output.txt")
        stderr = DockerCodeExecutionBackend._get_container_file_static(container, f"{workspace_dir}/error.txt")

        # Get created files
        try:
            bits, _ = container.get_archive(workspace_dir, encode_stream=True)
            bio = io.BytesIO()
            for chunk in bits:
                bio.write(chunk)
            bio.seek(0)
            files = str(bio.getvalue())
        except Exception:
            files = ""

        if status_code == 0:
            return CodeExecutionOutput(
                message="success",
                stdout=stdout,
                stderr=stderr,
                files=files,
            )

        # Create a more descriptive error message
        error_message = f"Program exited with status code {status_code}"
        if stderr:
            error_message = f"{error_message}\n\nError details:\n{stderr}"

        return CodeExecutionOutput(
            message="error",
            error=error_message,
            stdout=stdout,
            stderr=stderr,
            files=files,
        )

    @staticmethod
    def _get_container_file_static(container, file_path: str) -> str:
        """Static version of _get_container_file for use in processes."""
        try:
            bits, _ = container.get_archive(file_path)
            bio = io.BytesIO()
            for chunk in bits:
                bio.write(chunk)
            bio.seek(0)

            with tarfile.open(fileobj=bio) as tar:
                member = tar.next()
                f = tar.extractfile(member)
                if f:
                    return f.read().decode("utf-8")
                return ""

        except docker.errors.NotFound:
            return ""

    async def shutdown(self):
        """Cleanup resources including process pool."""
        self.process_pool.shutdown(wait=True)
        self.client.close()
        logger.debug("Shutting down Docker backend and process pool")



from typing import Any, Dict, List

from pydantic import BaseModel, Field

"""A generic request model for LLM API requests.

Attributes:
    model: The name/identifier of the LLM model to use.
    messages: List of message dictionaries forming the conversation history.
    response_format: Optional json schema of a Pydantic model class for structured output validation.
        None indicates non-structured (str) output is expected.
    original_row: The source data being processed. The original row as a dictionary.
    original_row_idx: The index of the original row in the dataset.
    multimodal_prompt: Flag indicating if the prompt contains multimodal content.
"""


class GenericRequest(BaseModel):
    """A generic request model for LLM API requests."""

    model: str
    messages: List[Dict[str, Any]]
    response_format: Dict[str, Any] | None = None
    original_row: Dict[str, Any]
    original_row_idx: int
    generation_params: Dict[str, Any] = Field(default_factory=dict)
    is_multimodal_prompt: bool | None = False



"""Curator: Bespoke Labs Synthetic Data Generation Library."""

import inspect
import os
from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Type, TypeVar

from datasets import Dataset
from pydantic import BaseModel
from xxhash import xxh64

from bespokelabs.curator.client import Client
from bespokelabs.curator.constants import _CURATOR_DEFAULT_CACHE_DIR, _INTERNAL_PROMPT_KEY
from bespokelabs.curator.db import MetadataDB
from bespokelabs.curator.llm.prompt_formatter import PromptFormatter
from bespokelabs.curator.log import add_file_handler, logger
from bespokelabs.curator.request_processor._factory import _RequestProcessorFactory
from bespokelabs.curator.request_processor.config import BackendParamsType

if TYPE_CHECKING:
    from dataset import Dataset

T = TypeVar("T")
_DictOrBaseModel = Dict[str, Any] | BaseModel


class LLM:
    """Interface for prompting LLMs."""

    response_format: Type[BaseModel] | None = None
    return_completions_object: bool = False

    def prompt(self, input: _DictOrBaseModel) -> _DictOrBaseModel:
        """Prompt the LLM.

        Args:
            input: The input row used to construct the prompt.

        Returns:
            The prompt to send to the LLM. Can follow the following formats:

            1. A string, corresponding to a single user prompt, e.g.
            The string "Write a poem about love" will be converted
            to [{"role": "user", "content": "Write a poem about love"}]

            2. A list of dictionaries, corresponding to a list of messages, e.g.
            The list [{"role": "user", "content": "Write a poem about love"},
            {"role": "assistant", "content": "Here is a poem about love"}]
        """
        return input

    def parse(self, input: _DictOrBaseModel, response: _DictOrBaseModel) -> _DictOrBaseModel:
        """Parse the response from the LLM and combine it with the input.

        Args:
            input: The input row used to construct the prompt
            response: The response from the LLM

        Returns:
            The parsed output row that combines the input and response,
        """
        if isinstance(response, str):
            return {"response": response}
        elif isinstance(response, BaseModel):
            return response.model_dump()
        return response

    def __init__(
        self,
        model_name: str,
        response_format: Type[BaseModel] | None = None,
        batch: bool = False,
        backend: Optional[str] = None,
        generation_params: dict | None = None,
        backend_params: BackendParamsType | None = None,
    ):
        """Initialize a LLM.

        Args:
            model_name: The name of the LLM to use
            response_format: A Pydantic model specifying the
                response format from the LLM
            batch: Whether to use batch processing
            backend: The backend to use ("openai", "litellm", or "vllm"). If None, will be auto-determined
            generation_params: Additional parameters to pass to the generation API
            backend_params: Dictionary parameters for request processing
                    - max_retries: The maximum number of retries to use for the LLM
                    - require_all_responses: Whether to require all responses
                    - base_url: Optional base URL for API endpoint
                    - request_timeout: Timeout in seconds for each request

            Other backend params:
                - Online:
                    - max_requests_per_minute: Maximum number of requests per minute for rate limiting
                    - max_tokens_per_minute: Maximum number of tokens per minute for rate limiting (combined input and output)
                    - max_input_tokens_per_minute: Maximum number of input tokens per minute for rate limiting
                    - max_output_tokens_per_minute: Maximum number of output tokens per minute for rate limiting
                    - seconds_to_pause_on_rate_limit: Number of seconds to pause when rate limited

                - Batch:
                    - batch_size: The size of the batch to use, only used if batch is True
                    - batch_check_interval: The interval to check for batch completions, only used if batch is True
                    - delete_successful_batch_files: Whether to delete successful batch files, only used if batch is True
                    - delete_failed_batch_files: Whether to delete failed batch files, only used if batch is True

                - Offline:
                    - tensor_parallel_size: The tensor parallel size to use for the VLLM backend
                    - enforce_eager: Whether to enforce eager execution for the VLLM backend
                    - max_model_length: The maximum model length to use for the VLLM backend
                    - max_tokens: The maximum tokens to use for the VLLM backend
                    - min_tokens: The minimum tokens to use for the VLLM backend
                    - gpu_memory_utilization: The GPU memory utilization to use for the VLLM backend
                    - batch_size: The size of the batch to use, only used if batch is True
        """
        generation_params = generation_params or {}

        if response_format is not None:
            self.response_format = response_format

        self.prompt_formatter = PromptFormatter(
            model_name=model_name,
            prompt_func=self.prompt,
            parse_func=self.parse,
            response_format=self.response_format,
            generation_params=_remove_none_values(generation_params),
        )
        self.batch_mode = batch

        self._request_processor = _RequestProcessorFactory.create(
            params=backend_params,
            model_name=model_name,
            batch=batch,
            response_format=response_format,
            backend=backend,
            generation_params=generation_params,
            return_completions_object=self.return_completions_object,
        )

    def _hash_fingerprint(self, dataset_hash, disable_cache):
        if disable_cache:
            fingerprint = xxh64(os.urandom(8)).hexdigest()
        else:
            # Get the source code of the prompt and parse methods
            prompt_func_hash = _get_function_hash(self.prompt_formatter.prompt_func)

            fingerprint_str = "_".join(
                [
                    str(dataset_hash),
                    str(prompt_func_hash),
                    str(self.prompt_formatter.model_name),
                    str(self.prompt_formatter.response_format.model_json_schema() if self.prompt_formatter.response_format else "text"),
                    str(self.batch_mode),
                ]
            )

            if self.prompt_formatter.generation_params:
                generation_params_str = str(sorted(self.prompt_formatter.generation_params.items()))
                fingerprint_str += f"_{generation_params_str}"

            fingerprint = xxh64(fingerprint_str.encode("utf-8")).hexdigest()
            logger.debug(f"Curator Cache Fingerprint String: {fingerprint_str}")
            logger.debug(f"Curator Cache Fingerprint: {fingerprint}")

        return fingerprint

    def __call__(
        self,
        dataset: Optional[Iterable] = None,
        working_dir: str = None,
        batch_cancel: bool = False,
    ) -> "Dataset":
        """Apply structured completions in parallel to a dataset using specified model and prompts.

        Args:
            dataset (Iterable): A dataset consisting of a list of items to apply completions
            working_dir (str): The working directory to save the requests.jsonl, responses.jsonl, and dataset.arrow files.
            batch_cancel (bool): Whether to cancel the batch if it is running
        Returns:
            Iterable: A list of structured outputs from the completions
        """
        # We convert from iterable to Dataset because Dataset has random access via row_idx
        if dataset:
            dataset = _convert_to_dataset(dataset)

        if working_dir is None:
            curator_cache_dir = os.environ.get(
                "CURATOR_CACHE_DIR",
                os.path.expanduser(_CURATOR_DEFAULT_CACHE_DIR),
            )
        else:
            curator_cache_dir = working_dir

        dataset_hash = dataset._fingerprint if dataset is not None else xxh64("").hexdigest()

        disable_cache = os.getenv("CURATOR_DISABLE_CACHE", "").lower() in ["true", "1"]
        fingerprint = self._hash_fingerprint(dataset_hash, disable_cache)

        metadata_db_path = os.path.join(curator_cache_dir, "metadata.db")
        metadata_db = MetadataDB(metadata_db_path)
        self._request_processor.viewer_client = Client()

        # Get the source code of the prompt function
        prompt_func_source = _get_function_source(self.prompt_formatter.prompt_func)
        if self.prompt_formatter.parse_func is not None:
            parse_func_source = _get_function_source(self.prompt_formatter.parse_func)
        else:
            parse_func_source = ""

        metadata_dict = {
            "timestamp": datetime.now().isoformat(),
            "dataset_hash": dataset_hash,
            "prompt_func": prompt_func_source,
            "parse_func": parse_func_source,
            "model_name": self.prompt_formatter.model_name,
            "response_format": (str(self.prompt_formatter.response_format.model_json_schema()) if self.prompt_formatter.response_format else "text"),
            "run_hash": fingerprint,
            "batch_mode": self.batch_mode,
        }

        run_cache_dir = os.path.join(curator_cache_dir, fingerprint)
        os.makedirs(run_cache_dir, exist_ok=True)
        add_file_handler(run_cache_dir)

        existing_session_id = metadata_db.get_existing_session_id(metadata_dict["run_hash"])
        existing_viewer_sync = metadata_db.check_existing_hosted_sync(metadata_dict["run_hash"])
        if not existing_viewer_sync and existing_session_id:
            session_id = self._request_processor.viewer_client.create_session(metadata_dict)
        else:
            session_id = self._request_processor.viewer_client.create_session(metadata_dict, session_id=existing_session_id)

        metadata_dict["session_id"] = session_id
        metadata_dict["is_hosted_viewer_synced"] = False
        metadata_db.store_metadata(metadata_dict)

        if batch_cancel:
            from bespokelabs.curator.request_processor.batch.openai_batch_request_processor import OpenAIBatchRequestProcessor

            if not isinstance(self._request_processor, OpenAIBatchRequestProcessor):
                raise ValueError("batch_cancel can only be used with batch mode")

            dataset = self._request_processor.cancel_batches(
                working_dir=run_cache_dir,
            )
        else:
            parse_func_hash = _get_function_hash(self.prompt_formatter.parse_func)
            dataset = self._request_processor.run(
                dataset=dataset,
                working_dir=run_cache_dir,
                parse_func_hash=parse_func_hash,
                prompt_formatter=self.prompt_formatter,
            )

        if existing_session_id is not None and existing_viewer_sync is False:
            msg = (
                f"There was a previous run with the same run hash ({metadata_dict['run_hash']}) without the CURATOR_VIEWER flag enabled, "
                "and CURATOR_VIEWER flag is enabled for this run. This means that the Curator Viewer is potentially inconsistent with local data."
                "Pushing the full dataset to the Curator Viewer to ensure full consistency."
            )
            if self._request_processor.viewer_client.hosted:
                logger.warning(msg)
                from bespokelabs.curator.utils import push_to_viewer

                push_to_viewer(dataset, session_id=session_id)

        if self._request_processor.viewer_client.hosted:
            metadata_db.update_sync_viewer_flag(metadata_dict["run_hash"], True)
        return dataset


def _get_function_hash(func) -> str:
    # setting recursion limit to avoid random recursion limit error
    # todo: understand why this is happening and fix it
    import sys

    sys.setrecursionlimit(10000)

    """Get a hash of a function's source code."""
    if func is None:
        return xxh64("").hexdigest()

    # Remove parameter annotations to avoid dill complaining about pickling
    # pydantic BaseModel: https://github.com/bespokelabsai/curator/issues/229.
    # For class/instance methods, get the underlying function
    if hasattr(func, "__func__"):
        func = func.__func__

    # Clear annotations if they exist
    if hasattr(func, "__annotations__"):
        func.__annotations__ = {}

    file = BytesIO()

    from datasets.utils._dill import Pickler

    try:
        Pickler(file, recurse=True).dump(func)
    except TypeError:
        logger.debug("Failed to recursive pickle function, trying non-recursive")
        Pickler(file, recurse=False).dump(func)

    return xxh64(file.getvalue()).hexdigest()


def _get_function_source(func) -> str:
    """Get the source code of a function.

    Purpose of this function is that during Python interpreter (REPL),
    `inspect.getsource` will fail with an OSError because functions defined in the
    interpreter don't have an associated source file. We have to use this wrapper
    to gracefully handle this case.
    """
    try:
        return inspect.getsource(func)
    except OSError:
        return ""


def _remove_none_values(d: dict) -> dict:
    """Remove all None values from a dictionary."""
    return {k: v for k, v in d.items() if v is not None}


def _is_message_list(list: list) -> bool:
    """Check if a list is a list of messages."""
    return all(isinstance(item, dict) and "role" in item and "content" in item for item in list)


def _convert_to_dataset(iterable: Iterable) -> "Dataset":
    """Convert an iterable to a Dataset.

    The prompt is expected to be a prompt string or a list of messages.
    It will be stored with the key '__internal_prompt' internally.
    """
    if isinstance(iterable, str) or _is_message_list(iterable):
        # A single string or list of messages is converted to a dataset with a single row
        dataset = Dataset.from_list([{_INTERNAL_PROMPT_KEY: iterable}])
    elif not isinstance(iterable, Dataset) and iterable is not None:
        # Wrap the iterable in a generator, the prompt is expected to be a prompt string or a list of messages
        def wrapped_iterable():
            for input in iterable:
                if isinstance(input, str) or _is_message_list(input):
                    yield {_INTERNAL_PROMPT_KEY: input}
                else:
                    yield input

        dataset = Dataset.from_generator(wrapped_iterable)
    else:
        dataset = iterable

    return dataset



"""Base class for online request processors that make real-time API calls.

This module provides the core functionality for making API requests in real-time,
handling rate limiting, retries, and concurrent processing.
"""

import asyncio
import datetime
import json
import os
import time
import typing as t
from abc import ABC, abstractmethod
from collections import Counter, deque
from dataclasses import dataclass, field

import aiofiles
import aiohttp

from bespokelabs.curator.llm.prompt_formatter import PromptFormatter
from bespokelabs.curator.log import logger
from bespokelabs.curator.request_processor import _DEFAULT_COST_MAP
from bespokelabs.curator.request_processor.base_request_processor import BaseRequestProcessor
from bespokelabs.curator.request_processor.config import OnlineRequestProcessorConfig
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop
from bespokelabs.curator.status_tracker.online_status_tracker import OnlineStatusTracker, TokenLimitStrategy
from bespokelabs.curator.types.generic_request import GenericRequest
from bespokelabs.curator.types.generic_response import GenericResponse
from bespokelabs.curator.types.prompt import _MultiModalPrompt
from bespokelabs.curator.types.token_usage import _TokenUsage

_MAX_OUTPUT_MVA_WINDOW = 50


@dataclass
class APIRequest:
    """Stores an API request's inputs, outputs, and other metadata.

    Attributes:
        task_id: Unique identifier for the request
        generic_request: The generic request object to be processed
        api_specific_request: The request formatted for the specific API
        attempts_left: Number of retry attempts remaining
        result: List to store results/errors from attempts
        prompt_formatter: Formatter for prompts and responses
        created_at: Timestamp when request was created
    """

    task_id: int
    generic_request: GenericRequest
    api_specific_request: dict
    attempts_left: int
    result: list = field(default_factory=list)
    prompt_formatter: PromptFormatter = field(default=None)
    created_at: datetime.datetime = field(default_factory=datetime.datetime.now)


class BaseOnlineRequestProcessor(BaseRequestProcessor, ABC):
    """Abstract base class for online request processors that make real-time API calls.

    This class handles rate limiting, retries, parallel processing and other common
    functionality needed for making real-time API requests.

    Args:
        config: Configuration object containing settings for the request processor
    """

    def __init__(self, config: OnlineRequestProcessorConfig):
        """Initialize the BaseOnlineRequestProcessor."""
        super().__init__(config)

        defaults = _DEFAULT_COST_MAP["online"]["default"]["ratelimit"]
        self.token_limit_strategy = TokenLimitStrategy.default
        self.manual_max_requests_per_minute = config.max_requests_per_minute
        self.manual_max_tokens_per_minute = config.max_tokens_per_minute
        self.default_max_requests_per_minute = defaults["max_requests_per_minute"]
        self.default_max_concurrent_requests = defaults["max_concurrent_requests"]
        self.default_max_tokens_per_minute = defaults["max_tokens_per_minute"][self.token_limit_strategy.value]
        self.header_based_max_requests_per_minute = None
        self.header_based_max_tokens_per_minute = None

        self.manual_max_concurrent_requests = config.max_concurrent_requests
        self.header_based_max_concurrent_requests = None

        # The rich.Console used for the status tracker, only set for testing
        self._tracker_console = None
        self._output_tokens_window = deque(maxlen=_MAX_OUTPUT_MVA_WINDOW)
        self._semaphore = None
        if self.max_concurrent_requests is not None:
            self._semaphore = asyncio.Semaphore(t.cast(int, self.max_concurrent_requests))

    @property
    def backend(self) -> str:
        """Backend property."""
        return "base"

    @property
    def compatible_provider(self) -> str:
        """Compatible provider property."""
        return "base"

    def validate_config(self):
        """Validate online request processor configuration.

        Ensures that configuration parameters are set correctly.

        Raises:
            ValueError: If configuration parameters are invalid
        """
        pass

    def _unpack_multimodal(self, request: GenericRequest) -> GenericRequest:
        messages = request.messages
        if request.is_multimodal_prompt is True and self._multimodal_prompt_supported is False:
            raise TypeError(f"Multimodal prompts are not supported by this model {self.config.model}.")

        if request.is_multimodal_prompt is False:
            return request

        unpacked_messages = []
        for message in messages:
            # Load message content as multimodal prompt
            content = _MultiModalPrompt.model_validate(message["content"])
            content = self._handle_multi_modal_prompt(content)
            message["content"] = content
            unpacked_messages.append(message)

        request.messages = unpacked_messages
        return request

    @abstractmethod
    def file_upload_limit_check(self, base64_image: str) -> None:
        """Check if the image size is within the allowed limit."""
        pass

    def _handle_multi_modal_prompt(self, message):
        content = []
        texts = message.texts

        def _format_multimodal(data, mime_type="image/png"):
            """Format multimodal prompt data for API request."""
            mime_type = mime_type or "image/png"
            if data.url and not data.is_local:
                return {"type": "image_url", "image_url": {"url": data.url}}
            else:
                base64_content = data.serialize()
                self.file_upload_limit_check(base64_content)

                content = {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:{mime_type};base64,{base64_content}",
                    },
                }
                if "image" in mime_type:
                    content["image_url"].update({"detail": data.detail})
                return content

        for text in texts:
            content.append({"type": "text", "text": text})
        for image in message.images:
            content.append(_format_multimodal(image, mime_type=image.mime_type))
        for file in message.files:
            content.append(_format_multimodal(file, mime_type=file.mime_type))

        return content

    @property
    def max_concurrent_requests(self) -> int | None:
        """Gets the maximum concurrent requests rate limit.

        Returns the manually set limit if available, falls back to header-based limit,
        or uses default value as last resort.
        """
        if self.manual_max_concurrent_requests:
            logger.info(f"Manually set max_concurrent_requests to {self.manual_max_concurrent_requests}")
            return self.manual_max_concurrent_requests

        elif self.header_based_max_concurrent_requests:
            logger.info(f"Automatically set max_concurrent_requests to {self.header_based_max_concurrent_requests}")
            return self.header_based_max_concurrent_requests
        else:
            return None

    @property
    def max_requests_per_minute(self) -> int:
        """Gets the maximum requests per minute rate limit.

        Returns the manually set limit if available, falls back to header-based limit,
        or uses default value as last resort.
        """
        if self.manual_max_requests_per_minute:
            logger.info(f"Manually set max_requests_per_minute to {self.manual_max_requests_per_minute}")
            return self.manual_max_requests_per_minute
        elif self.header_based_max_requests_per_minute:
            logger.info(f"Automatically set max_requests_per_minute to {self.header_based_max_requests_per_minute}")
            return self.header_based_max_requests_per_minute
        else:
            logger.warning(
                f"No manual max_requests_per_minute set, and headers based detection failed, using default value of {self.default_max_requests_per_minute}"
            )
            return self.default_max_requests_per_minute

    @property
    def max_tokens_per_minute(self) -> int:
        """Gets the maximum tokens per minute rate limit.

        Returns the manually set limit if available, falls back to header-based limit,
        or uses default value as last resort.
        """
        if self.manual_max_tokens_per_minute:
            logger.info(f"Manually set max_tokens_per_minute to {self.manual_max_tokens_per_minute}")
            return self.manual_max_tokens_per_minute
        elif self.header_based_max_tokens_per_minute:
            logger.info(f"Automatically set max_tokens_per_minute to {self.header_based_max_tokens_per_minute}")
            return self.header_based_max_tokens_per_minute
        else:
            logger.warning(
                f"No manual max_tokens_per_minute set, and headers based detection failed, using default value of {self.default_max_tokens_per_minute}"
            )
            return self.default_max_tokens_per_minute

    @abstractmethod
    def estimate_total_tokens(self, messages: list) -> _TokenUsage:
        """Estimate total tokens for a request.

        Args:
            messages: List of messages to estimate token count for

        Returns:
            Estimated total number of tokens
        """
        pass

    @abstractmethod
    def estimate_output_tokens(self) -> int:
        """Estimate output tokens for a request.

        Returns:
            Estimated number of output tokens
        """
        pass

    @abstractmethod
    def create_api_specific_request_online(self, generic_request: GenericRequest) -> dict:
        """Create an API-specific request body from a generic request body.

        Args:
            generic_request: The generic request to convert

        Returns:
            API-specific request dictionary
        """
        pass

    def completion_cost(self, response):
        """Calculate the cost of a completion response using litellm.

        Args:
            response: The completion response to calculate cost for

        Returns:
            Calculated cost of the completion
        """
        # Calculate cost using litellm
        cost = self._cost_processor.cost(completion_response=response)
        return cost

    def requests_to_responses(
        self,
        generic_request_files: list[str],
    ) -> None:
        """Process multiple request files and generate corresponding response files.

        Args:
            generic_request_files: List of request files to process
        """
        for request_file in generic_request_files:
            response_file = request_file.replace("requests_", "responses_")
            run_in_event_loop(
                self.process_requests_from_file(
                    generic_request_filepath=request_file,
                    response_file=response_file,
                )
            )

    async def cool_down_if_rate_limit_error(self, status_tracker: OnlineStatusTracker) -> None:
        """Pause processing if a rate limit error is detected.

        Args:
            status_tracker: Tracker containing rate limit status
        """
        seconds_to_pause_on_rate_limit = self.config.seconds_to_pause_on_rate_limit
        seconds_since_rate_limit_error = time.time() - status_tracker.time_of_last_rate_limit_error
        remaining_seconds_to_pause = seconds_to_pause_on_rate_limit - seconds_since_rate_limit_error
        if remaining_seconds_to_pause > 0:
            logger.warn(f"Pausing for {int(remaining_seconds_to_pause)} seconds")
            await asyncio.sleep(remaining_seconds_to_pause)
            status_tracker.last_update_time = time.time()

    def free_capacity(self, tracker, tokens):
        """Free blocked capacity."""

    async def process_requests_from_file(
        self,
        generic_request_filepath: str,
        response_file: str,
    ) -> None:
        """Processes API requests in parallel, throttling to stay under rate limits.

        Args:
            generic_request_filepath: Path to file containing requests
            response_file: Path where the response data will be saved
        """
        # Initialize trackers
        queue_of_requests_to_retry: asyncio.Queue[APIRequest] = asyncio.Queue()
        status_tracker = OnlineStatusTracker(
            token_limit_strategy=self.token_limit_strategy,
            max_requests_per_minute=self.max_requests_per_minute,
            max_tokens_per_minute=self.max_tokens_per_minute,
            compatible_provider=self.compatible_provider,
            viewer_client=self._viewer_client,
        )

        # Resume if a response file exists
        completed_request_ids, completed_parsed_responses = self.validate_existing_response_file(response_file)

        # Count total requests
        status_tracker.num_tasks_already_completed = len(completed_request_ids)
        status_tracker.num_parsed_responses = completed_parsed_responses
        status_tracker.total_requests = self.total_requests
        status_tracker.model = self.prompt_formatter.model_name
        status_tracker.start_tracker(self._tracker_console)

        # Use higher connector limit for better throughput
        tcp_limit = self.max_concurrent_requests if status_tracker.max_requests_per_minute is None else status_tracker.max_requests_per_minute
        # Update session status to inprogress
        await self._viewer_client.session_inprogress()
        connector = aiohttp.TCPConnector(limit=10 * tcp_limit)
        async with aiohttp.ClientSession(connector=connector) as session:
            async with aiofiles.open(generic_request_filepath) as file:
                pending_requests = []
                async for line in file:
                    if self._semaphore:
                        await self._semaphore.acquire()

                    generic_request = GenericRequest.model_validate_json(line)

                    if generic_request.original_row_idx in completed_request_ids:
                        continue

                    # Unpack multimodal prompts
                    generic_request = self._unpack_multimodal(generic_request)

                    request = APIRequest(
                        task_id=status_tracker.num_tasks_started,
                        generic_request=generic_request,
                        api_specific_request=self.create_api_specific_request_online(generic_request),
                        attempts_left=self.config.max_retries,
                        prompt_formatter=self.prompt_formatter,
                    )

                    if status_tracker.max_tokens_per_minute is not None:
                        token_estimate = self.estimate_total_tokens(request.generic_request.messages)
                    else:
                        token_estimate = None

                    # Wait for capacity if needed
                    while not status_tracker.has_capacity(token_estimate):
                        await asyncio.sleep(0.1)

                    # Wait for rate limits cool down if needed
                    await self.cool_down_if_rate_limit_error(status_tracker)

                    # Consume capacity before making request
                    status_tracker.consume_capacity(token_estimate)
                    task = asyncio.create_task(
                        self.handle_single_request_with_retries(
                            request=request,
                            session=session,
                            retry_queue=queue_of_requests_to_retry,
                            response_file=response_file,
                            status_tracker=status_tracker,
                            blocked_capacity=token_estimate,
                        )
                    )
                    pending_requests.append(task)

                    status_tracker.num_tasks_started += 1
                    status_tracker.num_tasks_in_progress += 1

            if pending_requests:
                await asyncio.gather(*pending_requests)

            # Process any remaining retries in the queue
            pending_retries = set()
            while not queue_of_requests_to_retry.empty() or pending_retries:
                # Process new items from the queue if we have capacity
                if self._semaphore:
                    await self._semaphore.acquire()

                if not queue_of_requests_to_retry.empty():
                    retry_request = await queue_of_requests_to_retry.get()
                    if status_tracker.max_tokens_per_minute is not None:
                        token_estimate = self.estimate_total_tokens(retry_request.generic_request.messages)
                    else:
                        token_estimate = None

                    attempt_number = self.config.max_retries - retry_request.attempts_left
                    logger.debug(
                        f"Retrying request {retry_request.task_id} "
                        f"(attempt #{attempt_number} of {self.config.max_retries})"
                        f"Previous errors: {retry_request.result}"
                    )

                    # Wait for capacity if needed
                    while not status_tracker.has_capacity(token_estimate):
                        await asyncio.sleep(0.1)

                    # Wait for rate limits cool down if needed
                    await self.cool_down_if_rate_limit_error(status_tracker)

                    # Consume capacity before making request
                    status_tracker.consume_capacity(token_estimate)

                    task = asyncio.create_task(
                        self.handle_single_request_with_retries(
                            request=retry_request,
                            session=session,
                            retry_queue=queue_of_requests_to_retry,
                            response_file=response_file,
                            status_tracker=status_tracker,
                            blocked_capacity=token_estimate,
                        )
                    )
                    pending_retries.add(task)

                # Wait for some tasks to complete
                if pending_retries:
                    done, pending_retries = await asyncio.wait(pending_retries, timeout=0.1)

        status_tracker.stop_tracker()
        await self.viewer_client.session_completed()

        # Log final status
        logger.info(f"Processing complete. Results saved to {response_file}")
        logger.info(f"Status tracker: {status_tracker}")

        if status_tracker.num_tasks_failed > 0:
            logger.warning(f"{status_tracker.num_tasks_failed} / {status_tracker.num_tasks_started} requests failed. Errors logged to {response_file}.")

    def _free_capacity(self, status_tracker: OnlineStatusTracker, used_capacity: "_TokenUsage", blocked_capacity: "_TokenUsage"):
        if status_tracker.max_tokens_per_minute is not None:
            status_tracker.free_capacity(used_capacity, blocked_capacity)

    async def handle_single_request_with_retries(
        self,
        request: APIRequest,
        session: aiohttp.ClientSession,
        retry_queue: asyncio.Queue,
        response_file: str,
        status_tracker: OnlineStatusTracker,
        blocked_capacity: "_TokenUsage",
    ) -> None:
        """Common wrapper for handling a single request with error handling and retries.

        This method implements the common try/except logic and retry mechanism,
        while delegating the actual API call to call_single_request.

        Args:
            request: The request to process
            session: Async HTTP session
            retry_queue: Queue for failed requests
            response_file: Path where the response data will be saved
            status_tracker: Tracks request status
            blocked_capacity: Blocked token capacity
        """
        try:
            # Estimate tokens before making request
            generic_response = None
            token_estimate: _TokenUsage = blocked_capacity or self.estimate_total_tokens(request.generic_request.messages)

            # Add new estimate to projection (pre_request=True indicates new estimate)
            status_tracker.update_cost_projection(token_estimate, pre_request=True)

            generic_response = await self.call_single_request(
                request=request,
                session=session,
                status_tracker=status_tracker,
            )

            if generic_response.finish_reason in self.config.invalid_finish_reasons:
                logger.debug(
                    f"Invalid finish_reason {generic_response.finish_reason}."
                    " Raw response {generic_response.raw_response} "
                    "for request {generic_response.raw_request}"
                )
                # Update cost projection with actual usage but mark as failed
                status_tracker.update_stats(generic_response.token_usage, generic_response.response_cost)
                raise ValueError(f"finish_reason was {generic_response.finish_reason}")

            # Allows us to retry on responses that don't match the response format
            self.prompt_formatter.response_to_response_format(generic_response.response_message)

        except Exception as e:
            status_tracker.num_other_errors += 1
            request.result.append(e)

            if generic_response and generic_response.token_usage is not None:
                used_tokens: _TokenUsage = _TokenUsage(input=generic_response.token_usage.input, output=generic_response.token_usage.output)
                status_tracker.update_cost_projection(used_tokens)
            else:
                status_tracker.update_cost_projection(None)

            if request.attempts_left > 0:
                request.attempts_left -= 1
                logger.warning(
                    f"Encountered '{e.__class__.__name__}: {e}' during attempt "
                    f"{self.config.max_retries - request.attempts_left} of {self.config.max_retries} "
                    f"while processing request {request.task_id}"
                )
                retry_queue.put_nowait(request)
            else:
                error_counts = Counter(str(err) for err in request.result)
                formatted_errors = [f"{error}(x{count})" for error, count in error_counts.items()]
                logger.error(
                    f"Request {request.task_id} failed permanently after exhausting all {self.config.max_retries} retry attempts. "
                    f"Errors: {formatted_errors}"
                )

                generic_response = GenericResponse(
                    response_message=None,
                    response_errors=formatted_errors,
                    raw_request=request.api_specific_request,
                    raw_response=None,
                    generic_request=request.generic_request,
                    created_at=request.created_at,
                    finished_at=datetime.datetime.now(),
                )
                await self.append_generic_response(status_tracker, generic_response, response_file)
                status_tracker.num_tasks_in_progress -= 1
                status_tracker.num_tasks_failed += 1
            return
        else:
            self._add_output_token_moving_window(generic_response.token_usage.output)
        finally:
            if self._semaphore:
                self._semaphore.release()

        # Save response in the base class
        await self.append_generic_response(status_tracker, generic_response, response_file)

        status_tracker.num_tasks_in_progress -= 1
        status_tracker.num_tasks_succeeded += 1

        # Update cost projection with actual usage
        used_tokens: _TokenUsage = _TokenUsage(input=generic_response.token_usage.input, output=generic_response.token_usage.output)
        # Update stats with actual usage
        status_tracker.update_stats(generic_response.token_usage, generic_response.response_cost)
        status_tracker.update_cost_projection(used_tokens)

        # Free the extra capacity blocked before request with actual consumed capacity.
        self._free_capacity(status_tracker, used_tokens, blocked_capacity)

    def _add_output_token_moving_window(self, tokens):
        self._output_tokens_window.append(tokens)

    def _output_tokens_moving_average(self):
        return sum(self._output_tokens_window) / (len(self._output_tokens_window) or _MAX_OUTPUT_MVA_WINDOW)

    @abstractmethod
    async def call_single_request(
        self,
        request: APIRequest,
        session: aiohttp.ClientSession,
        status_tracker: OnlineStatusTracker,
    ) -> GenericResponse:
        """Make a single API request without error handling.

        This method should implement the actual API call logic
        without handling retries or errors.

        Args:
            request: Request to process
            session: Async HTTP session
            status_tracker: Tracks request status

        Returns:
            The response from the API call
        """
        pass

    async def append_generic_response(self, status_tracker: OnlineStatusTracker, data: GenericResponse, filename: str) -> None:
        """Append a response to a jsonl file with async file operations.

        Args:
            status_tracker: Tracker containing request status
            data: Response data to append
            filename: File to append to
        """
        raw_response = data.response_message
        responses = self._process_response(data)
        if not responses:
            return
        data.parsed_response_message = responses
        data.response_message = raw_response
        data_dump = data.model_dump()
        json_string = json.dumps(data_dump, default=str)
        async with aiofiles.open(filename, "a") as f:
            await f.write(json_string + "\n")
        logger.debug(f"Successfully appended response to {filename}")
        filename = os.path.basename(filename).split(".")[0]
        idx = status_tracker.num_parsed_responses
        status_tracker.num_parsed_responses = idx + len(responses)
        await self.viewer_client.stream_response(json_string, idx)



import asyncio
import json
import os
from abc import abstractmethod
from collections import Counter
from typing import Optional

import aiofiles
from litellm import model_cost

from bespokelabs.curator.log import logger
from bespokelabs.curator.request_processor.base_request_processor import BaseRequestProcessor
from bespokelabs.curator.request_processor.config import BatchRequestProcessorConfig
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop
from bespokelabs.curator.status_tracker.batch_status_tracker import BatchStatusTracker
from bespokelabs.curator.types.generic_batch import GenericBatch, GenericBatchRequestCounts, GenericBatchStatus
from bespokelabs.curator.types.generic_request import GenericRequest
from bespokelabs.curator.types.generic_response import GenericResponse
from bespokelabs.curator.types.token_usage import _TokenUsage


class BaseBatchRequestProcessor(BaseRequestProcessor):
    """Abstract base class for processing batched API requests.

    This class provides the core functionality for submitting, tracking, and managing
    batch requests across different LLM API providers. It handles file operations,
    request tracking, batch status management, and response processing.

    The batch processing workflow:
    1. Load or initialize batch status tracker
    2. Submit requests in batches to API
    3. Monitor batch processing status
    4. Download and process completed batches
    5. Handle failures and retries

    Attributes:
        working_dir (str): Directory for storing batch-related files.
        prompt_formatter (PromptFormatter): Formatter for structuring prompts.
        tracker (BatchStatusTracker): Tracks status of submitted batches.
        batch_objects_file (str): Path to file storing batch objects.
        semaphore (asyncio.Semaphore): Controls concurrent batch operations.

    Note:
        Subclasses must implement abstract methods to provide API-specific
        functionality while maintaining consistent batch processing behavior.
    """

    def __init__(self, config: BatchRequestProcessorConfig):
        """Initialize the batch request processor.

        Args:
            config: Configuration object containing batch processing parameters.
        """
        super().__init__(config)
        self._tracker_console = None  # Add this line to store console for testing

    @property
    def backend(self) -> str:
        """Backend property."""
        return "base"

    @property
    def compatible_provider(self) -> str:
        """Compatible provider property."""
        return self.backend

    def requests_to_responses(self, generic_request_files: list[str]) -> None:
        """Process multiple request files using batch API operations.

        Orchestrates the complete batch processing workflow:
        1. Validates batch size limits
        2. Initializes concurrent operation controls
        3. Submits requests in batches
        4. Monitors and processes batch results

        Args:
            generic_request_files: List of paths to files containing requests.

        Raises:
            ValueError: If batch size exceeds API limits.
            RuntimeError: If batch processing fails or no successful responses.

        Side Effects:
            - Creates batch tracking files in working directory
            - Updates progress bars for batch submission and processing
            - Generates response files for completed batches
        """
        self.semaphore = asyncio.Semaphore(self.max_concurrent_batch_operations)
        self._batch_objects_file_lock = asyncio.Lock()
        self.batch_objects_file = os.path.join(self.working_dir, "batch_objects.jsonl")

        self._attempt_loading_batch_status_tracker(generic_request_files)

        run_in_event_loop(self.submit_batches_from_request_files(generic_request_files))
        run_in_event_loop(self.poll_and_process_batches())
        self.tracker.stop_tracker()

    @property
    @abstractmethod
    def max_requests_per_batch(self) -> int:
        """Maximum number of requests allowed in a single batch.

        This property must be implemented by subclasses to specify their
        API-specific batch size limits.

        Returns:
            int: Maximum number of requests that can be included in one batch.

        Note:
            This limit is enforced during batch creation to prevent API errors.
        """
        pass

    @property
    @abstractmethod
    def max_bytes_per_batch(self) -> int:
        """Maximum size in bytes allowed for a single batch.

        This property must be implemented by subclasses to specify their
        API-specific batch size limits in bytes.

        Returns:
            int: Maximum allowed size of a batch in bytes.

        Note:
            This limit is enforced during batch file creation to prevent API errors.
        """
        pass

    @property
    @abstractmethod
    def max_concurrent_batch_operations(self) -> int:
        """Maximum number of concurrent batch operations allowed.

        This property must be implemented by subclasses to specify their
        API-specific concurrency limits for batch operations.

        Returns:
            int: Maximum number of batch operations that can run concurrently.

        Note:
            This limit is enforced via semaphore during batch operations.
        """
        pass

    @abstractmethod
    async def submit_batch(self, requests: list[dict], metadata: Optional[dict] = None) -> GenericBatch:
        """Submit a batch of requests to the API provider.

        Args:
            requests: List of API-specific request dictionaries.
            metadata: Optional metadata to associate with the batch.

        Returns:
            GenericBatch: Standardized batch object with submission details.

        Note:
            Implementation must use self.semaphore for concurrency control.
        """
        pass

    @abstractmethod
    async def retrieve_batch(self, batch: GenericBatch) -> GenericBatch:
        """Retrieve current status of a submitted batch.

        Args:
            batch: The batch object to check status for.

        Returns:
            GenericBatch: Updated batch object with current status.

        Note:
            Implementation must use self.semaphore for concurrency control.
        """
        pass

    @abstractmethod
    async def cancel_batch(self, batch: GenericBatch) -> GenericBatch:
        """Cancel a running batch job.

        Args:
            batch: The batch object to cancel.

        Returns:
            GenericBatch: Updated batch object after cancellation attempt.

        Note:
            Implementation must use self.semaphore for concurrency control.
        """
        pass

    @abstractmethod
    async def download_batch(self, batch: GenericBatch) -> str | None:
        """Download results of a completed batch.

        Args:
            batch: The completed batch object to download.

        Returns:
            str | None: Raw response content if successful, None if failed.

        Note:
            Implementation should handle API-specific result formats.
        """
        pass

    @abstractmethod
    def parse_api_specific_response(
        self,
        raw_response: dict,
        generic_request: GenericRequest,
        batch: GenericBatch,
    ) -> GenericResponse:
        """Parse API-specific response into standardized format.

        Args:
            raw_response: Raw response dictionary from API.
            generic_request: Original generic request object.
            batch: Batch object containing context information.

        Returns:
            GenericResponse: Standardized response object.

        Note:
            Should handle both successful and failed responses.
        """
        pass

    @abstractmethod
    def create_api_specific_request_batch(self, generic_request: GenericRequest) -> dict:
        """Convert generic request to API-specific format.

        Args:
            generic_request: Standardized request object.

        Returns:
            dict: API-specific request dictionary.

        Note:
            Must format request according to API provider requirements.
        """
        pass

    @abstractmethod
    def parse_api_specific_batch_object(self, batch: object, request_file: str | None = None) -> GenericBatch:
        """Convert API-specific batch object to generic format.

        Args:
            batch: API-specific batch object.
            request_file: Optional path to associated request file.

        Returns:
            GenericBatch: Standardized batch object.
        """
        pass

    @abstractmethod
    def parse_api_specific_request_counts(self, request_counts: object, request_file: Optional[str] = None) -> GenericBatchRequestCounts:
        """Convert API-specific request counts to generic format.

        Args:
            request_counts: API-specific request count object.
            request_file: Path to associated request file.

        Returns:
            GenericBatchRequestCounts: Standardized request count object.
        """
        pass

    def validate_config(self):
        """Validate batch request processor configuration.

        Ensures that configuration parameters are set correctly for batch processing.

        Raises:
            ValueError: If configuration parameters are invalid
        """
        if self.config.batch_size > self.max_requests_per_batch:
            raise ValueError(
                f"batch_size {self.config.batch_size} is greater than the maximum of "
                f"{self.max_requests_per_batch:,} requests per batch that {self.__class__.__name__} supports. "
                f"Please set your batch_size to be less than or equal to {self.max_requests_per_batch:,}."
            )

    def _attempt_loading_batch_status_tracker(self, request_files: set[str]):
        """Load existing batch status tracker or create new one.

        Args:
            request_files: Set of paths to request files to track.

        Side Effects:
            - Loads tracker from batch_objects_file if it exists
            - Creates new tracker if file doesn't exist
            - Sets self.tracker with loaded/created tracker
        """
        if os.path.exists(self.batch_objects_file):
            with open(self.batch_objects_file, "r") as f:
                self.tracker = BatchStatusTracker.model_validate_json(f.read())
                self.tracker.viewer_client = self._viewer_client
            logger.info(f"Loaded existing tracker from {self.batch_objects_file}")
        else:
            self.tracker = BatchStatusTracker(
                unsubmitted_request_files=set(request_files),
                viewer_client=self._viewer_client,
            )

        self.tracker.model = self.prompt_formatter.model_name
        self.tracker.n_total_requests = self.total_requests
        self.set_model_cost()

    def set_model_cost(self):
        """Set cost information for the current model."""
        # Set cost information if available
        if self.prompt_formatter.model_name in model_cost:
            # Batch requests are 50% cheaper
            self.tracker.input_cost_per_million = (model_cost[self.prompt_formatter.model_name]["input_cost_per_token"] * 1_000_000) * 0.5
            self.tracker.output_cost_per_million = (model_cost[self.prompt_formatter.model_name]["output_cost_per_token"] * 1_000_000) * 0.5
        else:
            from bespokelabs.curator.cost import external_model_cost

            self.tracker.input_cost_per_million = (
                external_model_cost(self.prompt_formatter.model_name, provider=self.compatible_provider, completion_window=self.config.completion_window)[
                    "input_cost_per_token"
                ]
                * 1_000_000
            )
            self.tracker.output_cost_per_million = (
                external_model_cost(self.prompt_formatter.model_name, provider=self.compatible_provider, completion_window=self.config.completion_window)[
                    "output_cost_per_token"
                ]
                * 1_000_000
            )

        # Start the tracker with the console from constructor
        self.tracker.start_tracker(self._tracker_console)

    async def update_batch_objects_file(self):
        """Update batch objects file with current tracker state.

        Side Effects:
            - Writes current tracker state to batch_objects_file
            - Uses file lock to prevent concurrent writes
        """
        async with self._batch_objects_file_lock:
            with open(self.batch_objects_file, "w") as f:
                f.write(self.tracker.model_dump_json())

    def create_batch_file(self, api_specific_requests: list[dict]) -> str:
        """Create a batch file from API-specific requests.

        Validates request count and file size against API limits before creating
        the batch file content.

        Args:
            api_specific_requests: List of API-specific request dictionaries.

        Returns:
            str: Encoded file content ready for API upload.

        Raises:
            ValueError: If batch exceeds request count or size limits.

        Side Effects:
            - Logs debug information about batch file size
        """
        n_requests = len(api_specific_requests)
        if n_requests > self.max_requests_per_batch:
            raise ValueError(
                f"Batch file contains {n_requests:,} requests, "
                f"which is more than the maximum of {self.max_requests_per_batch:,} requests per batch that {self.__class__.__name__} supports. "
                f"Preventing batch submission. Please reduce `batch_size`."
            )

        # Join requests with newlines and encode to bytes for upload
        file_content = "\n".join(json.dumps(r) for r in api_specific_requests).encode()
        file_content_size = len(file_content)
        logger.debug(f"Batch file content size: {file_content_size / (1024 * 1024):.2f} MB ({file_content_size:,} bytes)")
        if file_content_size > self.max_bytes_per_batch:
            raise ValueError(
                f"Batch file content size {file_content_size:,} bytes "
                f"is greater than the maximum of {self.max_bytes_per_batch:,} bytes per batch that {self.__class__.__name__} supports. "
                f"Please reduce your batch size or request content size (via prompt_func and response_format)."
            )
        return file_content

    async def submit_batch_from_request_file(
        self,
        request_file: str,
        completed_request_ids: set[int],
    ):
        """Submit a batch of requests from a file.

        Reads requests from file, converts them to API-specific format,
        and submits them as a batch.

        Args:
            request_file: Path to file containing request data.
            completed_request_ids: Set of request IDs that have already been completed
                and should be skipped.

        Side Effects:
            - Updates batch submission progress bar
            - Updates tracker with submitted batch status
            - Creates batch metadata with request file path
            - Updates batch objects file
        """
        metadata = {"request_file": request_file}
        requests = self.requests_from_generic_request_file(request_file, completed_request_ids)
        batch = await self.submit_batch(requests, metadata)
        self.tracker.mark_as_submitted(batch, len(requests))
        await self.update_batch_objects_file()

    def requests_from_generic_request_file(self, request_file: str, completed_request_ids: set[int]) -> list[dict]:
        """Reads and converts generic requests from a file into API-specific request format.

        Args:
            request_file (str): Path to the file containing generic requests in JSONL format.
            completed_request_ids (set[int]): Set of request IDs that already have responses, these will be skipped
        Returns:
            list[dict]: API-specific request bodies ready for submission.

        Side Effects:
            - Reads from request file on disk
        """
        api_specific_requests = []

        with open(request_file, "r") as file:
            for line in file:
                request = GenericRequest.model_validate_json(line.strip())
                if request.original_row_idx in completed_request_ids:
                    continue
                api_specific_request = self.create_api_specific_request_batch(request)
                api_specific_requests.append(api_specific_request)

        return api_specific_requests

    async def generic_response_file_from_responses(self, responses: list[dict], batch: GenericBatch) -> str | None:
        """Process API responses and create generic response file.

        Converts API-specific responses to GenericResponse objects and writes them
        to a response file. Handles successful and failed responses, including
        token usage tracking and cost calculation.

        Args:
            responses: List of raw API response dictionaries.
            batch: Batch object containing request metadata.

        Returns:
            str | None: Path to created response file, or None if creation failed.

        Side Effects:
            - Creates response file from request file name
            - Writes GenericResponse objects in JSONL format
            - Calculates costs with batch discount
            - Handles failed requests with error details
            - Updates tracker with token usage and cost stats
        """
        request_file = batch.request_file
        request_dir = os.path.dirname(request_file)
        request_filename = os.path.basename(request_file)
        response_filename = request_filename.replace("requests_", "responses_")
        response_file = os.path.join(request_dir, response_filename)

        # Load generic requests from request file
        generic_request_map = {}
        with open(request_file, "r") as f:
            for line in f:
                generic_request = GenericRequest.model_validate_json(line)
                generic_request_map[generic_request.original_row_idx] = generic_request

        # Track total token usage and cost for this batch
        total_token_usage = _TokenUsage(input=0, output=0)
        total_cost = 0.0

        # appending allows for the resubmitted resumed batch

        stream_response_tasks = []
        invalid_finish_responses = []
        failed_processed_responses = []
        async with aiofiles.open(response_file, "a") as f:
            for raw_response in responses:
                request_idx = int(raw_response["custom_id"])
                generic_request = generic_request_map[request_idx]
                generic_response = self.parse_api_specific_response(raw_response, generic_request, batch)

                if generic_response.finish_reason in self.config.invalid_finish_reasons:
                    invalid_finish_responses.append({"request_id": request_idx, "finish_reason": generic_response.finish_reason})
                    continue

                processed_responses = self._process_response(generic_response)
                generic_response.parsed_response_message = processed_responses
                if processed_responses is None:
                    failed_processed_responses.append(request_idx)
                    continue

                # Write response to file
                response_dump = generic_response.model_dump(mode="json")
                r = json.dumps(response_dump, default=str)
                await f.write(r + "\n")

                # Update token and cost totals
                if generic_response.token_usage:
                    total_token_usage.input += generic_response.token_usage.input
                    total_token_usage.output += generic_response.token_usage.output
                if generic_response.response_cost:
                    total_cost += generic_response.response_cost

                # Stream responses to viewer client
                idx = self.tracker.num_parsed_responses
                self.tracker.num_parsed_responses = idx + len(processed_responses)
                stream_response_tasks.append(self.viewer_client.stream_response(json.dumps(response_dump), idx))

        await asyncio.gather(*stream_response_tasks)
        if failed_processed_responses:
            logger.warning(f"Batch {batch.id} has {len(failed_processed_responses)} failed responses due to parse function errors.")

        if invalid_finish_responses:
            logger.warning(f"Batch {batch.id} has {len(invalid_finish_responses)} invalid finish responses. Please check the logs above for details.")
            invalid_finish_reasons = dict(Counter([response["finish_reason"] for response in invalid_finish_responses]))
            logger.warning(f"Invalid finish responses: {invalid_finish_reasons}")

        # Update tracker with token usage and cost stats
        self.tracker.update_token_and_cost(total_token_usage, total_cost)
        await self.viewer_client.session_completed()
        return response_file

    async def submit_batches_from_request_files(
        self,
        request_files: set[str],
    ):
        """Submit multiple request files as batches to API.

        Manages the complete batch submission workflow including tracking,
        progress monitoring, and concurrent submission of multiple files.

        Args:
            request_files: Set of paths to request files to process.

        Side Effects:
            - Loads or creates batch status tracker
            - Updates tracker with batch statuses
            - Creates and updates batch submission progress bar
            - Submits batches concurrently using asyncio
            - Updates batch objects file
        """
        tasks = []

        # Update session status to inprogress
        await self._viewer_client.session_inprogress()

        # check existing response files for resuming
        for batch in self.tracker.downloaded_batches.values():
            response_file = batch.request_file.replace("requests_", "responses_")
            completed_request_ids, completed_parsed_responses = self.validate_existing_response_file(response_file)
            self.tracker.num_parsed_responses += completed_parsed_responses
            n_total_batch_requests = self.read_metadata_file(batch.request_file).get("num_jobs")
            if len(completed_request_ids) < n_total_batch_requests:
                tasks.append(self.submit_batch_from_request_file(batch.request_file, completed_request_ids))

        # submit full batches of unsubmitted request files
        for request_file in self.tracker.unsubmitted_request_files:
            tasks.append(self.submit_batch_from_request_file(request_file, set()))

        # exit early if no batches to submit
        if len(tasks) == 0:
            return

        await asyncio.gather(*tasks)
        assert self.tracker.unsubmitted_request_files == set()

    async def check_batch_status(self, batch: GenericBatch) -> GenericBatch | None:
        """Check current status of a batch job.

        Retrieves current batch status from API and updates tracking information.
        Handles batch completion detection and request count updates.

        Args:
            batch: The batch object to check status for.

        Returns:
            GenericBatch | None: Updated batch object if found, None if not found.

        Side Effects:
            - Updates tracker with current batch status
            - Updates request completion counts
            - Logs batch status and request counts
            - Marks completed batches as finished
        """
        async with self.semaphore:
            batch = await self.retrieve_batch(batch)
            if batch is not None:
                self.tracker.update_submitted(batch)

                n_succeeded_requests = batch.request_counts.succeeded
                n_failed_requests = batch.request_counts.failed
                n_total_requests = batch.request_counts.total

                logger.debug(
                    f"Batch {batch.id} status: {batch.raw_status} requests: "
                    f"{n_succeeded_requests}/{n_failed_requests}/{n_total_requests} "
                    "succeeded/failed/total"
                )

                if batch.status == GenericBatchStatus.FINISHED.value:
                    logger.debug(f"Batch {batch.id} finished with status: {batch.raw_status}")
                    self.tracker.mark_as_finished(batch)
                    await self.update_batch_objects_file()

    async def poll_and_process_batches(self) -> None:
        """Monitor and process batches until completion.

        Continuously polls batch status and downloads results when complete.
        Manages batch lifecycle including status checks, downloads, and error handling.

        Returns:
            None

        Raises:
            RuntimeError: If no batches complete successfully.

        Side Effects:
            - Creates and updates request progress bar
            - Updates batch tracker state
            - Downloads and processes completed batches
            - Creates response files for completed batches
            - Logs progress and status information
            - Sleeps between status checks
        """
        # loop until all batches have been returned
        all_response_files = []
        while self.tracker.n_submitted_batches + self.tracker.n_finished_batches > 0:
            # check batch status also updates the tracker
            status_tasks = [self.check_batch_status(batch) for batch in self.tracker.submitted_batches.values()]
            await asyncio.gather(*status_tasks)
            await self.update_batch_objects_file()

            download_tasks = [self.download_batch_to_response_file(batch) for batch in self.tracker.finished_batches.values()]
            # Failed downloads return None and print any errors that occurred
            all_response_files.extend(await asyncio.gather(*download_tasks))

            logger.debug(
                f"Batches returned: {self.tracker.n_finished_or_downloaded_batches:,}/{self.tracker.n_total_batches:,} "
                f"Requests completed: {self.tracker.n_finished_or_downloaded_succeeded_requests:,}/{self.tracker.n_total_requests:,}"
            )

            self.tracker.update_display()
            if self.tracker.n_submitted_batches + self.tracker.n_finished_batches > 0:
                logger.debug(f"Sleeping for {self.config.batch_check_interval} seconds...")
                await asyncio.sleep(self.config.batch_check_interval)

        response_files = filter(None, all_response_files)
        await self.viewer_client.close()
        if self.tracker.n_downloaded_batches == 0 or not response_files:
            raise RuntimeError(f"None of the submitted batches completed successfully. Please check the logs above and {self.web_dashboard} for errors.")

    async def download_batch_to_response_file(self, batch: GenericBatch) -> str | None:
        """Download and process completed batch results."""
        file_content = await self.download_batch(batch)

        if file_content is None:
            return None

        # Write responses to file and update stats
        response_file = await self.generic_response_file_from_responses(file_content, batch)

        logger.debug(f"Batch {batch.id} written to {response_file}")

        if self.config.delete_successful_batch_files:
            await self.delete_file(batch.input_file_id, self.semaphore)
            await self.delete_file(batch.output_file_id, self.semaphore)

        # Update tracker with downloaded batch
        self.tracker.mark_as_downloaded(batch)
        await self.update_batch_objects_file()

        return response_file

    async def cancel_batches(self):
        """Cancel all currently submitted batches.

        Attempts to cancel all batches that are currently in submitted state.
        Handles cases where no batches are submitted.

        Side Effects:
            - Attempts to cancel all submitted batches concurrently
            - Logs warning if no batches to cancel
            - Updates batch status through cancel_batch calls
        """
        if self.tracker.n_submitted_batches == 0:
            logger.warning("No batches to be cancelled, but cancel_batches=True.")
            return
        tasks = [self.cancel_batch(batch) for batch in self.tracker.submitted_batches.values()]
        await asyncio.gather(*tasks)



from dataclasses import dataclass
from typing import Callable

from bespokelabs.curator.code_executor.types import CodeExecutionRequest, CodeExecutionRequestParams


@dataclass
class CodeFormatter:
    """Formatter for the code execution backend."""

    code: Callable
    code_input: Callable
    code_output: Callable
    execution_params: CodeExecutionRequestParams

    def create_code_execution_request(self, row: dict, idx: int, execution_directory: str) -> CodeExecutionRequest:
        """Format the request object based off of `LLM` attributes.

        Args:
            row: Input data to format into a prompt
            idx: Index of the row in the dataset
            execution_directory: Directory to create the temp file in

        Returns:
            CodeExecutionRequest object containing the formatted request
        """
        return CodeExecutionRequest(
            code=self.code(row),
            code_input=self.code_input(row),
            execution_params=self.execution_params,
            original_row=row,
            original_row_idx=idx,
            execution_directory=execution_directory,
        )



import logging
import logging.handlers
import os
import sys

from rich.console import Console
from rich.logging import RichHandler

LOG_FORMAT = "%(asctime)s | %(levelname)-8s | %(module)s:%(lineno)d - %(message)s"
ROOT_LOG_LEVEL = logging.DEBUG

_CONSOLE = Console(stderr=True)


class Logger:
    """Curator Logger class with handlers."""

    _instance = None

    def __new__(cls):
        """Python new method."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._init_logger()
        return cls._instance

    def _init_logger(self):
        self.logger = logging.getLogger("curator")
        self.logger.setLevel(ROOT_LOG_LEVEL)
        IN_COLAB = "google.colab" in sys.modules
        if IN_COLAB:
            self.logger.propagate = False
        if not self.logger.handlers:
            rich_handler = RichHandler(console=_CONSOLE)
            rich_handler.setLevel(logging.INFO)

            self.logger.addHandler(rich_handler)

    def get_logger(self, name):
        """Get logger instance."""
        return self.logger.getChild(name)


logger = Logger().get_logger(__name__)


def add_file_handler(log_dir):
    """Create a file handler and attach it to logger."""
    global logger
    log_file = os.path.join(log_dir, "curator.log")
    formatter = logging.Formatter(LOG_FORMAT)
    file_handler = logging.handlers.RotatingFileHandler(log_file, maxBytes=5 * 1024 * 1024, backupCount=5)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)



"""Base module for request processing functionality."""

import asyncio
import functools
import glob
import json
import os
import resource
import typing as t
from abc import ABC, abstractmethod
from math import ceil
from typing import TYPE_CHECKING, List, Optional

import aiofiles
import pyarrow
from pydantic import BaseModel, ValidationError

from bespokelabs.curator.constants import _CACHE_MSG, _INTERNAL_PROMPT_KEY
from bespokelabs.curator.cost import cost_processor_factory
from bespokelabs.curator.file_utilities import count_lines
from bespokelabs.curator.hf_card_template import HUGGINGFACE_CARD_TEMPLATE
from bespokelabs.curator.llm.prompt_formatter import PromptFormatter
from bespokelabs.curator.log import logger
from bespokelabs.curator.request_processor.config import BatchRequestProcessorConfig, RequestProcessorConfig
from bespokelabs.curator.request_processor.event_loop import run_in_event_loop
from bespokelabs.curator.types.generic_response import GenericResponse

if TYPE_CHECKING:
    from datasets import Dataset


class BaseRequestProcessor(ABC):
    """Base class for all request processors.

    Provides core functionality for processing requests through LLM APIs, including:
    - File descriptor limit management
    - Request file creation and caching
    - Response processing and dataset generation
    - Error handling and validation
    """

    def __init__(self, config: RequestProcessorConfig):
        """Initialize the request processor.

        Args:
            config: Configuration object containing processing parameters
        """
        # Increase the number of open file descriptors to avoid "Too many open files" errors
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        desired_limit = min(10_000_000, hard)
        logger.debug(f"Adjusting file descriptor limit from {soft} to {desired_limit} (hard limit: {hard})")
        resource.setrlimit(resource.RLIMIT_NOFILE, (desired_limit, hard))
        self._viewer_client = None
        self.config = config
        self._cost_processor = cost_processor_factory(config=self.config, backend=self.backend)

    @property
    @abstractmethod
    def backend(self) -> str:
        """Backend property."""
        return "base"

    @abstractmethod
    def validate_config(self):
        """Validate request processor configuration.

        Ensures that configuration parameters are set correctly.

        Raises:
            ValueError: If configuration parameters are invalid
        """
        pass

    @abstractmethod
    def requests_to_responses(self, generic_request_files: list[str]) -> None:
        """Process request files and generate responses.

        Args:
            generic_request_files: List of paths to request files to process
        """
        pass

    @property
    def viewer_client(self):
        """Return the viewer client for the request processor."""
        return self._viewer_client

    @viewer_client.setter
    def viewer_client(self, client):
        """Set the viewer client for the request processor."""
        self._viewer_client = client

    def check_structured_output_support(self) -> bool:
        """Check if the model supports structured output.

        Returns:
            bool: True if structured output is supported, False otherwise
        """
        return True

    def run(
        self,
        dataset: "Dataset",
        working_dir: str,
        parse_func_hash: str,
        prompt_formatter: PromptFormatter,
    ) -> "Dataset":
        """Uses the API to completing the specific map by calling the LLM.

        Args:
            dataset: Dataset that is being mapped over
            working_dir: Working directory to save files (requests.jsonl, responses.jsonl, dataset.arrow)
            parse_func_hash: Hash of the parse function for caching
            prompt_formatter: Formatter for generating prompts from dataset rows

        Returns:
            Dataset: Completed dataset with LLM responses

        Raises:
            ValueError: If model doesn't support structured output but it's requested
        """
        self.prompt_formatter = prompt_formatter
        self.working_dir = working_dir
        self.total_requests = len(dataset) if dataset is not None else 1

        # load from already completed dataset
        output_dataset = self.attempt_loading_cached_dataset(parse_func_hash)
        if output_dataset is not None:
            return output_dataset
        logger.info(f"Running {self.__class__.__name__} completions with model: {self.config.model}")

        self.validate_config()
        self.prompt_formatter = prompt_formatter
        if self.prompt_formatter.response_format:
            if not self.check_structured_output_support():
                raise ValueError(f"Model {self.config.model} does not support structured output, response_format: {self.prompt_formatter.response_format}")
        generic_request_files = self.create_request_files(dataset)

        self.requests_to_responses(generic_request_files)
        return self.create_dataset_files(parse_func_hash)

    def _verify_existing_request_files(self, dataset: Optional["Dataset"]) -> List[int]:
        """Verify integrity of the cache and identify files needing regeneration.

        Checks that each request file has associated metadata and correct number of rows.

        Args:
            dataset: The dataset to create requests from

        Returns:
            List of indices for request files that need to be regenerated
        """
        if isinstance(self.config, BatchRequestProcessorConfig) and dataset is not None:
            expected_num_files = ceil(len(dataset) / self.config.batch_size)
        else:
            expected_num_files = 1

        try:
            incomplete_files = []
            for i in range(expected_num_files):
                req_f = os.path.join(self.working_dir, f"requests_{i}.jsonl")
                meta_f = os.path.join(self.working_dir, f"metadata_{i}.json")

                if not os.path.exists(req_f):
                    incomplete_files.append(i)
                    continue

                if not os.path.exists(meta_f):
                    logger.warning(f"Cache missing metadata file {meta_f} for request file {req_f}")
                    incomplete_files.append(i)
                    continue

                with open(req_f, "r") as f:
                    data = f.read()
                num_jobs = len(data.splitlines())

                with open(meta_f, "r") as f:
                    metadata = json.load(f)

                expected_num_jobs = metadata["num_jobs"]
                if num_jobs != expected_num_jobs:
                    logger.warning(f"Request file {req_f} has {num_jobs} jobs, but metadata file {meta_f} has {expected_num_jobs} jobs")
                    incomplete_files.append(i)
            return incomplete_files

        except Exception as e:
            logger.warning(f"Cache verification failed due to {e} - regenerating all request files.")
            incomplete_files = list(range(expected_num_files))
            return incomplete_files

    @property
    def _multimodal_prompt_supported(self) -> bool:
        return False

    def create_request_files(self, dataset: Optional["Dataset"]) -> list[str]:
        """Creates request files if they don't exist or uses existing ones.

        Args:
            dataset: The dataset to be processed

        Returns:
            List of paths to the request files
        """
        os.makedirs(self.working_dir, exist_ok=True)
        request_files = glob.glob(os.path.join(self.working_dir, "requests_*.jsonl"))

        # By default use existing requests in working_dir
        incomplete_files = self._verify_existing_request_files(dataset)

        if len(incomplete_files) == 0:
            logger.info(f"Using cached requests. {_CACHE_MSG}")
            # count existing jobs in file and print first job
            with open(request_files[0], "r") as f:
                # Count lines and store first job
                first_job = None
                num_jobs = 0
                for i, line in enumerate(f):
                    if i == 0:
                        first_job = json.loads(line)
                    num_jobs = i + 1

                if num_jobs > 0:
                    logger.debug(
                        f"There are {num_jobs} existing requests in {request_files[0]}\n"
                        f"Example request in {request_files[0]}:\n{json.dumps(first_job, default=str, indent=2)}"
                    )
                    return request_files

        # Create new requests file
        logger.info(f"Preparing request file(s) in {self.working_dir}")
        request_file = os.path.join(self.working_dir, "requests_0.jsonl")
        request_files = [request_file]

        metadata_file = os.path.join(self.working_dir, "metadata_0.json")
        metadata_files = [metadata_file]

        if dataset is None:
            with open(request_file, "w") as f:
                generic_request = self.prompt_formatter.create_generic_request(dict(), 0)  # noqa: C408
                if generic_request.is_multimodal_prompt is True:
                    assert self._multimodal_prompt_supported, "Requested processor does not support multimodal prompts."

                generic_request.generation_params = self.config.generation_params
                f.write(json.dumps(generic_request.model_dump(), default=str) + "\n")

            metadata_dict = {"num_jobs": 1}
            with open(metadata_file, "w") as f:
                f.write(json.dumps(metadata_dict, indent=4) + "\n")
            return request_files

        if isinstance(self.config, BatchRequestProcessorConfig):
            num_batches = ceil(len(dataset) / self.config.batch_size)
            request_files = [os.path.join(self.working_dir, f"requests_{i}.jsonl") for i in range(num_batches)]
            metadata_files = [os.path.join(self.working_dir, f"metadata_{i}.json") for i in range(num_batches)]

            async def create_all_request_files():
                tasks = [
                    self.acreate_request_file(
                        dataset,
                        request_files[i],
                        metadata_files[i],
                        start_idx=i * self.config.batch_size,
                    )
                    for i in range(num_batches)
                    if i in incomplete_files
                ]
                await asyncio.gather(*tasks)

            run_in_event_loop(create_all_request_files())
        else:
            run_in_event_loop(self.acreate_request_file(dataset, request_file, metadata_file))

        return request_files

    async def acreate_request_file(
        self,
        dataset: "Dataset",
        request_file: str,
        metadata_file: str,
        start_idx: int = 0,
    ) -> None:
        """Asynchronously create a request file and its metadata.

        Args:
            dataset: Dataset to create requests from
            request_file: Path to save request file
            metadata_file: Path to save metadata file
            start_idx: Starting index in dataset for this batch
        """
        if isinstance(self.config, BatchRequestProcessorConfig):
            end_idx = min(start_idx + self.config.batch_size, len(dataset))
            dataset = dataset.select(range(start_idx, end_idx))
        else:
            end_idx = len(dataset)

        # Check if we need to vary generation_params per row
        generation_params_per_row = "generation_params" in dataset.column_names
        if self.prompt_formatter.generation_params and generation_params_per_row:
            logger.warning("Found both default and row-level generation_params. Collided keys will follow values in row-level config.")
        async with aiofiles.open(request_file, "w") as f:
            for idx, dataset_row in enumerate(dataset):
                dataset_row_idx = idx + start_idx
                # Get the generic request from the map function
                request = self.prompt_formatter.create_generic_request(dataset_row, dataset_row_idx, generation_params_per_row)
                await f.write(json.dumps(request.model_dump(), default=str) + "\n")

        num_requests = end_idx - start_idx
        metadata_dict = {"num_jobs": num_requests}
        async with aiofiles.open(metadata_file, "w") as f:
            await f.write(json.dumps(metadata_dict, indent=4) + "\n")

        logger.info(f"Wrote {num_requests} requests to {request_file}.")

    def attempt_loading_cached_dataset(self, parse_func_hash: str) -> Optional["Dataset"]:
        """Attempt to load a cached dataset file.

        Args:
            parse_func_hash: Hash identifying the specific dataset

        Returns:
            Cached dataset if available and valid, None otherwise
        """
        dataset_file = os.path.join(self.working_dir, f"{parse_func_hash}.arrow")
        if os.path.exists(dataset_file):
            logger.debug(f"Loading dataset from {dataset_file}")
            try:
                logger.info(f"Using cached output dataset. {_CACHE_MSG}")
                return self._load_from_dataset_file(dataset_file)
            except pyarrow.lib.ArrowInvalid:
                os.remove(dataset_file)
                logger.warning(
                    f"Failed to load dataset from {dataset_file}, "
                    "which was likely corrupted by a failed previous run. "
                    "Deleted file and attempting to regenerate dataset from cached LLM responses."
                )

    def _process_response(self, data: GenericResponse) -> List | None:
        try:
            data.response_message = self.prompt_formatter.response_to_response_format(data.response_message)
        except (json.JSONDecodeError, ValidationError):
            logger.warning("Skipping response due to error parsing response message into response format")
            return

        # parse_func can return a single row or a list of rows
        responses = None
        if self.prompt_formatter.parse_func:
            try:
                if _INTERNAL_PROMPT_KEY in data.generic_request.original_row:
                    row = data.generic_request.original_row[_INTERNAL_PROMPT_KEY]
                else:
                    row = data.generic_request.original_row
                responses = self.prompt_formatter.parse_func(
                    row,
                    data.response_message,
                )
            except Exception as e:
                logger.warning(f"Skipping response due to error in `parse_func` :: {e}")
                return

            if not isinstance(responses, list):
                responses = [responses]
        else:
            # Convert response to dict before adding to dataset
            response_value = data.response_message
            if hasattr(response_value, "model_dump"):
                response_value = response_value.model_dump()
            elif hasattr(response_value, "__dict__"):
                response_value = response_value.__dict__
            responses = [{"response": response_value}]
        return responses

    def create_dataset_files(
        self,
        parse_func_hash: str,
    ) -> "Dataset":
        """Creates dataset from response files.

        Args:
            parse_func_hash: Hash identifying the dataset version

        Returns:
            Dataset containing processed responses

        Raises:
            ValueError: If no responses found or processing fails
        """
        responses_files = glob.glob(os.path.join(self.working_dir, "responses_*.jsonl"))
        if len(responses_files) == 0:
            raise ValueError(f"No responses files found in {self.working_dir}")

        error_help = (
            "Please check your `parse_func` is returning a valid row (dict) "
            "or list of rows (list of dicts) and re-run. "
            "Dataset will be regenerated from cached LLM responses."
        )

        # Process all response files
        total_responses_count = 0
        failed_responses_count = 0
        error_sample = []
        dataset_file = os.path.join(self.working_dir, f"{parse_func_hash}.arrow")
        from datasets.arrow_writer import ArrowWriter

        with ArrowWriter(path=dataset_file) as writer:
            for responses_file in responses_files:
                with open(responses_file, "r") as f_in:
                    for generic_response_string in f_in:
                        total_responses_count += 1
                        response = GenericResponse.model_validate_json(generic_response_string)

                        if response.response_errors is not None:
                            failed_responses_count += 1
                            if len(error_sample) < 10:
                                error_sample.append(str(response.response_errors))
                            continue

                        # TODO: Find a way to not process responses that have already been processed
                        # We cannot just check if parsed_response_message is not None because it could be from cached previous run
                        # response.
                        response.parsed_response_message = self._process_response(response)
                        if response.parsed_response_message is None:
                            failed_responses_count += 1
                            continue

                        for row in response.parsed_response_message:
                            if isinstance(row, BaseModel):
                                row = row.model_dump()

                            if not isinstance(row, dict):
                                os.remove(dataset_file)
                                raise ValueError(
                                    f"Got invalid row {row} of type {type(row)} from `parse_func`. This should be type <class 'dict'>. {error_help}"
                                )
                            if not row:
                                os.remove(dataset_file)
                                raise ValueError(f"Got empty row {row} from `parse_func`. {error_help}")
                            # Add the original row index to the row so that we can sort by it later.
                            row["__original_row_idx"] = response.generic_request.original_row_idx
                            writer.write(row)

            logger.info(f"Read {total_responses_count} responses.")
            error_sample_str = "\n".join(error_sample)
            error_sample_msg = f"Sample of the first {len(error_sample)} errors encountered: \n {error_sample_str}"
            if failed_responses_count == total_responses_count:
                writer.write({"error": "All requests failed"})
                writer.finalize()
                os.remove(dataset_file)
                raise ValueError(f"All requests failed. {error_sample_msg}")
            else:
                logger.info("Finalizing writer")
                writer.finalize()

            if failed_responses_count > 0:
                logger.warning(f"{failed_responses_count} requests failed.")
                if self.config.require_all_responses:
                    os.remove(dataset_file)
                    raise ValueError(f"Some requests failed and require_all_responses is True. {error_sample_msg}")

            # number of responses matches number of requests
            request_files = glob.glob(os.path.join(self.working_dir, "requests_*.jsonl"))
            n_requests = 0
            for request_file in request_files:
                n_requests += count_lines(request_file)

            if n_requests != total_responses_count:
                logger.warning(
                    f"{n_requests - total_responses_count} requests do not have responses. "
                    f"n_requests is {n_requests} and n_responses is {total_responses_count}"
                )
                if self.config.require_all_responses:
                    os.remove(dataset_file)
                    raise ValueError("Some requests do not have responses and require_all_responses is True.")

        return self._load_from_dataset_file(dataset_file)

    def _load_from_dataset_file(self, dataset_file: str) -> "Dataset":
        from datasets import Dataset

        d = Dataset.from_file(dataset_file)
        d = d.sort("__original_row_idx")
        d = d.remove_columns("__original_row_idx")

        push_to_hub = functools.partial(BaseRequestProcessor.push_to_hub, dataset=d, _push_to_hub=d.push_to_hub)
        d.push_to_hub = push_to_hub

        return d

    @staticmethod
    def push_to_hub(repo_id: str, dataset=None, _push_to_hub=None, **kwargs):
        """Push the dataset to the hub and create a dataset card."""
        from huggingface_hub import DatasetCard

        _push_to_hub(repo_id, **kwargs)
        card = DatasetCard(
            HUGGINGFACE_CARD_TEMPLATE.format(
                dataset_name=repo_id.split("/")[-1],
                repo_id=repo_id,
                sample=json.dumps(dataset[0], indent=4),
            )
        )
        card.push_to_hub(repo_id)

    def _get_validated_response(self, line: str) -> tuple[GenericResponse | None, bool]:
        """Check if a response is valid or has errors.

        Args:
            line: The line to process into a GenericResponse.

        Returns:
            response: The response if it is valid, None otherwise.
            is_valid: True if the response is valid, False otherwise.
        """
        response = None
        try:
            response = GenericResponse.model_validate_json(line)
            row_id = response.generic_request.original_row_idx
            if response.response_errors:
                logger.debug(f"Request {row_id} previously failed due to errors: {response.response_errors}, removing from output and will retry")
                return None, False
            if response.response_message is None:
                logger.debug(f"Request {row_id} previously failed due to no response. Removing from output and will retry.")
                return None, False
        except (json.JSONDecodeError, ValidationError):
            logger.warning(f"Skipping response due to error parsing line: {line}")
            return None, False

        return response, True

    def validate_existing_response_file(self, response_file: str) -> t.Union[set[int], int]:
        """Parse an existing response file to identify completed requests and removes failed requests.

        Args:
            response_file: Path to the response file to parse

        Returns:
            set[int]: Set of completed request IDs that were already successfully processed
            int: Number of completed parsed responses
        """
        if not os.path.exists(response_file):
            return set(), 0

        completed_request_ids = set()
        failed_request_ids = set()
        completed_parsed_responses = 0
        parsing_error_responses = 0
        logger.info(f"Resuming progress by reading existing file: {response_file}")
        logger.debug(f"Removing all failed requests from {response_file} so they can be retried")
        temp_filepath = response_file + ".temp"

        with open(response_file, "r") as input_file, open(temp_filepath, "w") as output_file:
            for line in input_file:
                response, is_valid = self._get_validated_response(line)
                if not response:
                    parsing_error_responses += 1
                    continue
                row_id = response.generic_request.original_row_idx
                if response.parsed_response_message:
                    completed_parsed_responses += len(response.parsed_response_message)
                if is_valid:
                    completed_request_ids.add(row_id)
                    output_file.write(line)
                else:
                    failed_request_ids.add(row_id)

        logger.info(
            f"Found {len(completed_request_ids)} successful requests and {len(failed_request_ids)} "
            f"previously failed requests and {parsing_error_responses} parsing errors in {response_file}"
        )
        os.replace(temp_filepath, response_file)

        return completed_request_ids, completed_parsed_responses

    def read_metadata_file(self, request_file: str) -> int:
        """Read the number of jobs from the metadata file.

        Args:
            request_file: Path to the request file to get metadata for

        Returns:
            int: Number of total batch requests

        Raises:
            ValueError: If metadata file is missing or invalid
        """
        metadata_file = request_file.replace("requests_", "metadata_").replace(".jsonl", ".json")

        if not os.path.exists(metadata_file):
            raise ValueError(f"Metadata file not found: {metadata_file}")

        try:
            with open(metadata_file, "r") as f:
                metadata = json.load(f)
                return metadata
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in metadata file: {metadata_file}. Delete cache directory 'rm -rf {self.working_dir}' and try again.") from e



import datetime
from dataclasses import field
from typing import Any, Dict, Optional

from pydantic import BaseModel


class CodeExecutionResult(BaseModel):
    """Result of the code execution backend."""

    stdout: str
    stderr: str
    exit_code: int


class CodeExecutionRequestParams(BaseModel):
    """Parameters for the code execution backend."""

    timeout: int = 10
    memory_limit: int = 1024 * 1024 * 1024


class CodeExecutionRequest(BaseModel):
    """Request to the code execution backend."""

    code: str
    code_input: str
    execution_params: Optional[CodeExecutionRequestParams] = None
    original_row: Optional[Dict[str, Any]] = None
    original_row_idx: Optional[int] = None
    execution_directory: Optional[str] = None


class CodeAPIRequest(BaseModel):
    """Request to the code execution backend."""

    task_id: Optional[int] = None
    execution_request: CodeExecutionRequest
    attempts_left: int
    code_formatter: Any
    created_at: datetime.datetime = field(default_factory=datetime.datetime.now)
    result: list = field(default_factory=list)


class CodeExecutionOutput(BaseModel):
    """Response from the code execution backend."""

    message: Optional[Dict[str, Any]] | str = None
    error: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    files: Optional[str] = None


class CodeExecutionResponse(BaseModel):
    """Response from the code execution backend."""

    exec_output: Optional[CodeExecutionOutput] = None
    code_api_request: Optional[CodeAPIRequest] = None
    created_at: datetime.datetime = field(default_factory=datetime.datetime.now)
    finished_at: datetime.datetime = field(default_factory=datetime.datetime.now)


class CodeExecutionBackendConfig(BaseModel):
    """Configuration for the code execution backend."""

    max_requests_per_minute: int = 10000
    max_retries: int = 3
    seconds_to_pause_on_rate_limit: int = 10
    base_url: Optional[str] = None
    docker_image: Optional[str] = None



HUGGINGFACE_CARD_TEMPLATE = """
---
language: en
license: mit
tags:
- curator
---

<a href="https://github.com/bespokelabsai/curator/">
 <img src="https://huggingface.co/datasets/bespokelabs/Bespoke-Stratos-17k/resolve/main/made_with_curator.png" alt="Made with Curator" width=200px>
</a>

## Dataset card for {dataset_name}

This dataset was made with [Curator](https://github.com/bespokelabsai/curator/).

## Dataset details

A sample from the dataset:

```python
{sample}
```

## Loading the dataset

You can load this dataset using the following code:

```python
from datasets import load_dataset

dataset = load_dataset("{repo_id}", split="default")
```

"""
